# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

"""NTLM Cryptographic Operations

Implements the various cryptographic operations that are defined in `MS-NLMP Crypto Operations`_. Some crypto
operations aren't implemented due to it being a simple operation in Python.

.. _MS-NLMP Crypto Operations
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/26c42637-9549-46ae-be2e-90f6f1360193
"""

import base64
import binascii
import hashlib
import hmac
import io
import re
import struct
import typing

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher

from spnego._ntlm_raw.des import DES
from spnego._ntlm_raw.md4 import md4
from spnego._ntlm_raw.messages import (
    FileTime,
    NegotiateFlags,
    NTClientChallengeV2,
    TargetInfo,
)

try:
    # cryptography 43.0.0 and later moved ARC4 to decrepit
    from cryptography.hazmat.decrepit.ciphers import algorithms
except ImportError:
    from cryptography.hazmat.primitives.ciphers import (  # type: ignore[no-redef]
        algorithms,
    )

# A user does not need to specify their actual plaintext password they can specify the LM and NT hash (from lmowfv1 and
# ntowfv2) in the form 'lm_hash_hex:nt_hash_hex'. This is still considered a plaintext pass as we can use it to build
# the LM and NT response but it's only usable for NTLM.
_NTLM_HASH_PATTERN = re.compile(r"^[a-fA-F0-9]{32}:[a-fA-F0-9]{32}$")


class RC4Handle:
    """RC4 class to wrap the underlying crypto function."""

    def __init__(self, key: bytes) -> None:
        self._key = key

        arc4 = algorithms.ARC4(self._key)
        cipher = Cipher(arc4, mode=None, backend=default_backend())
        self._handle = cipher.encryptor()

    def update(self, b_data: bytes) -> bytes:
        """Update the RC4 stream and return the encrypted/decrypted bytes."""
        return self._handle.update(b_data)

    def reset(self) -> None:
        """Reset's the cipher stream back to the original state."""
        arc4 = algorithms.ARC4(self._key)
        cipher = Cipher(arc4, mode=None, backend=default_backend())
        self._handle = cipher.encryptor()


def is_ntlm_hash(password: str) -> bool:
    return bool(_NTLM_HASH_PATTERN.match(password))


def compute_response_v1(
    flags: int,
    nt_hash: bytes,
    lm_hash: bytes,
    server_challenge: bytes,
    client_challenge: bytes,
    no_lm_response: bool = True,
) -> typing.Tuple[bytes, bytes, bytes]:
    """Compute NT and LM Response for NTLMv1.

    Computes the NT and LM Response for NTLMv1 messages. The response is dependent on the flags that were negotiated
    between the client and server.

    The pseudo-code for this function as documented under `NTLM v1 Authentication`_ is::

        Define ComputeResponse(NegFlg, ResponseKeyNT, ResponseKeyLM, CHALLENGE_MESSAGE.ServerChallenge,
            ClientChallenge, Time, ServerName) As

            If (User is set to "" AND Passwd is set to "")
                -- Special case for anonymous authentication
                Set NtChallengeResponseLen to 0
                Set NtChallengeResponseMaxLen to 0
                Set NtChallengeResponseBufferOffset to 0

                Set LmChallengeResponse to Z(1)
            ElseIf
                If (NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY flag is set in NegFlg)
                    Set NtChallengeResponse to DESL(ResponseKeyNT,
                        MD5(ConcatenationOf(CHALLENGE_MESSAGE.ServerChallenge, ClientChallenge))[0..7])

                    Set LmChallengeResponse to ConcatenationOf{ClientChallenge, Z(16)}
                Else
                    Set NtChallengeResponse to DESL(ResponseKeyNT, CHALLENGE_MESSAGE.ServerChallenge)

                    If (NoLMResponseNTLMv1 is TRUE)
                        Set LmChallengeResponse to NtChallengeResponse
                    Else
                        Set LmChallengeResponse to DESL(ResponseKeyLM, CHALLENGE_MESSAGE.ServerChallenge)
                    EndIf
                EndIf
            EndIf

        Set SessionBaseKey to MD4(NTOWF)

    Args:
        flags: The negotiated flags between the initiator and acceptor.
        nt_hash: The response key computed by :meth:`ntowfv1`.
        lm_hash: The response key computed by :meth:`lmowfv1`.
        server_challenge: The 8 byte nonce generated by the acceptor.
        client_challenge: The 8 byte nonce generated by the initiator.
        no_lm_response: Whether to compute (True) the `LmChallengeResponse` or not (False) when extended session
            security was not negotiated.

    Returns:
        Tuple[bytes, bytes, bytes]: Returns the NTChallengeResponse, LMChallengeResponse and KeyExchangeKey.

    .. _NTLM v1 Authentication:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/464551a8-9fc4-428e-b3d3-bc5bfb2e73a5
    """
    if flags & NegotiateFlags.extended_session_security:
        nt_response = desl(nt_hash, md5(server_challenge + client_challenge[:8]))
        lm_response = client_challenge + (b"\x00" * 16)

    else:
        nt_response = lm_response = desl(nt_hash, server_challenge)

        if not no_lm_response:
            lm_response = desl(lm_hash, server_challenge)

    session_base_key = md4(nt_hash)
    key_exchange_key = kxkey(flags, session_base_key, lm_hash, lm_response, server_challenge)

    return nt_response, lm_response, key_exchange_key


def compute_response_v2(
    nt_hash: bytes,
    server_challenge: bytes,
    client_challenge: bytes,
    time: FileTime,
    av_pairs: TargetInfo,
) -> typing.Tuple[bytes, bytes, bytes]:
    """Compute NT and LM Response for NTLMv2.

    Computes the NT and LM Response for NTLMv2 messages. The response is dependent on the flags that were negotiated
    between the client and server.

    The pseudo-code for this function as documented under `NTLM v2 Authentication`_ is::

        Define ComputeResponse(NegFlg, ResponseKeyNT, ResponseKeyLM, CHALLENGE_MESSAGE.ServerChallenge,
            ClientChallenge, Time, ServerName) As

            If (User is set to "" && Passwd is set to "")
                -- Special case for anonymous authentication
                Set NtChallengeResponseLen to 0
                Set NtChallengeResponseMaxLen to 0
                Set NtChallengeResponseBufferOffset to 0

                Set LmChallengeResponse to Z(1)
            Else
                Set temp to ConcatenationOf(Responserversion, HiResponserversion, Z(6), Time, ClientChallenge, Z(4),
                    ServerName, Z(4))

                Set NTProofStr to HMAC_MD5(ResponseKeyNT, ConcatenationOf(CHALLENGE_MESSAGE.ServerChallenge,temp))

                Set NtChallengeResponse to ConcatenationOf(NTProofStr, temp)

                Set LmChallengeResponse to ConcatenationOf(
                    HMAC_MD5(ResponseKeyLM, ConcatenationOf(CHALLENGE_MESSAGE.ServerChallenge, ClientChallenge)),
                    ClientChallenge)
            EndIf

        Set SessionBaseKey to HMAC_MD5(ResponseKeyNT, NTProofStr)

    Args:
        nt_hash: The response key computed by :meth:`ntwofv2`. The `ResponseKeyLM` is the same value so we only
            pass in the 1 key.
        server_challenge: The 8 byte nonce generated by the acceptor.
        client_challenge: The 8 byte nonce generated by the initiator.
        time: The FileTime to place in the NT hash.
        av_pairs: The TargetInfo AvPairs fields that are placed in the Authenticate message.

    Returns:
        Tuple[bytes, bytes, bytes]: Returns the NTChallengeResponse, LMChallengeResponse and KeyExchangeKey.

    .. _NTLM v2 Authentication:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/5e550938-91d4-459f-b67d-75d70009e3f3
    """
    temp = NTClientChallengeV2(time_stamp=time, client_challenge=client_challenge, av_pairs=av_pairs)
    b_temp = temp.pack() + b"\x00\x00\x00\x00"
    nt_proof_str = hmac_md5(nt_hash, server_challenge + b_temp)

    nt_response = nt_proof_str + b_temp
    lm_response = hmac_md5(nt_hash, server_challenge + client_challenge) + client_challenge
    session_base_key = hmac_md5(nt_hash, nt_proof_str)

    return nt_response, lm_response, session_base_key  # Is KeyExchangeKey in NTLMv2.


def crc32(m: bytes) -> bytes:
    """Simple wrapper function to generate a CRC32 checksum."""
    # We bitand to ensure the value is the same across all Python versions.
    return struct.pack("<I", binascii.crc32(m) & 0xFFFFFFFF)


def des(k: bytes, d: bytes) -> bytes:
    """DES encryption.

    Indicates the encryption of an 8-byte data item `d` with the 7-byte key `k` using the Data Encryption Standard
    (DES) algorithm in Electronic Codebook (ECB) mode. The result is 8 bytes in length ([FIPS46-2]).

    Args:
        k: The 7-byte key to use in the DES cipher.
        d: The 8-byte data block to encrypt.

    Returns:
        bytes: The encrypted data block.
    """
    return DES(DES.key56_to_key64(k)).encrypt(d)


def desl(k: bytes, d: bytes) -> bytes:
    """Encryption using the DES Long algorithm.

    Indicates the encryption of an 8-byte data item `d` with the 16-byte key `k` using the Data Encryption
    Standard Long (DESL) algorithm. The result is 24 bytes in length.

    `DESL(K, D)` as by MS-NLMP `DESL`_ is computed as follows::

        ConcatenationOf(
            DES(K[0..6], D),
            DES(K[7..13], D),
            DES(ConcatenationOf(K[14..15], Z(5)), D),
        );

    Args:
        k: The key to use for the DES cipher, will be truncated to 16 bytes and then padded to 21 bytes.
        d: The value to run through the DESL algorithm, will be truncated to 8 bytes.

    Returns:
        bytes: The output of the DESL algorithm.

    .. _DESL:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/26c42637-9549-46ae-be2e-90f6f1360193
    """
    k = k[:16].ljust(21, b"\x00")  # Key needs to be stripped at 16 characters and then padded to 21 chars.
    d = d[:8].ljust(8, b"\x00")  # Data need to be at most 8 bytes long.

    b_value = io.BytesIO()

    b_value.write(des(k[:7], d))
    b_value.write(des(k[7:14], d))
    b_value.write(des(k[14:], d))

    return b_value.getvalue()


def hmac_md5(key: bytes, data: bytes) -> bytes:
    """Simple wrapper function for a HMAC MD5 digest."""
    return hmac.new(key, data, digestmod=hashlib.md5).digest()


def kxkey(
    flags: int,
    session_base_key: bytes,
    lmowf: bytes,
    lm_response: bytes,
    server_challenge: bytes,
) -> bytes:
    """NTLM KXKEY function.

    The MS-NLMP `KXKEY`_ function used to derive the key exchange key for a security context. This is only for NTLMv1
    contexts as NTLMv2 just re-uses the session base key.

    Args:
        flags: The negotiate flags in the Challenge msg.
        session_base_key: The session base key from :meth:`compute_response_v1`.
        lmowf: The LM hash from :meth:`lmowfv1`.
        lm_response: The lm response from :meth:`compute_response_v1`.
        server_challenge: The server challenge in the Challenge msg.

    Returns:
        bytes: The derived key exchange key.

    .. _KXKEY:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/d86303b5-b29e-4fb9-b119-77579c761370
    """
    if flags & NegotiateFlags.extended_session_security:
        return hmac_md5(session_base_key, server_challenge + lm_response[:8])

    elif flags & NegotiateFlags.lm_key:
        b_data = lm_response[:8]
        return des(lmowf[:7], b_data) + des(lmowf[7:8] + b"\xbd\xbd\xbd\xbd\xbd\xbd", b_data)

    elif flags & NegotiateFlags.non_nt_session_key:
        return lmowf[:8] + b"\x00" * 8

    else:
        return session_base_key


def lmowfv1(password: str) -> bytes:
    """NTLMv1 LMOWFv1 function

    The Lan Manager v1 one way function as documented under `NTLM v1 Authentication`_.

    The pseudo-code for this function is::

        Define LMOWFv1(Passwd, User, UserDom) as
            ConcatenationOf(
                DES(UpperCase(Passwd)[0..6], "KGS!@#$%"),
                DES(UpperCase(Passwd)[7..13], "KGS!@#$%"),
            );

    Args:
        password: The password for the user.

    Returns:
        bytes: The LMv1 one way hash of the user's password.

    .. _NTLM v1 Authentication:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/464551a8-9fc4-428e-b3d3-bc5bfb2e73a5
    """
    if is_ntlm_hash(password):
        return base64.b16decode(password.split(":")[0].upper())

    # Fix the password to upper case and pad the length to exactly 14 bytes. While it is true LM only authentication
    # will fail if the password exceeds 14 bytes typically it is used in conjunction with the NTv1 hash which has no
    # such restrictions.
    b_password = password.upper().encode("utf-8", errors="surrogatepass").ljust(14, b"\x00")[:14]

    b_hash = io.BytesIO()
    for start, end in [(0, 7), (7, 14)]:
        b_hash.write(des(b_password[start:end], b"KGS!@#$%"))

    return b_hash.getvalue()


def md5(m: bytes) -> bytes:
    """Simple wrapper to generate a MD5 checksum."""
    return hashlib.md5(m).digest()


def ntowfv1(password: str) -> bytes:
    """NTLMv1 NTOWFv1 function

    The NT v1 one way function as documented under `NTLM v1 Authentication`_.

    The pseudo-code for this function is::

        Define NTOWFv1(Passwd, User, UserDom) as MD4(UNICODE(Passwd))

    Args:
        password: The password for the user.

    Returns:
        bytes: The NTv1 one way hash of the user's password.

    .. _NTLM v1 Authentication:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/464551a8-9fc4-428e-b3d3-bc5bfb2e73a5
    """
    if is_ntlm_hash(password):
        return base64.b16decode(password.split(":")[1].upper())

    return md4(password.encode("utf-16-le", errors="surrogatepass"))


def ntowfv2(username: str, nt_hash: bytes, domain_name: typing.Optional[str]) -> bytes:
    """NTLMv2 NTOWFv2 function

    The NT v2 one way function as documented under `NTLM v2 Authentication`_.

    The pseudo-code for this function is::

        Define NTOWFv2(Passwd, User, UserDom) as

            HMAC_MD5(MD4(UNICODE(Passwd)), UNICODE(ConcatenationOf(Uppercase(User), UserDom)))

    Args:
        username: The username.
        nt_hash: The NT hash from :meth:`ntowfv1`.
        domain_name: The optional domain name of the user.

    Returns:
        bytes: The NTv2 one way has of the user's credentials.

    .. _NTLM v2 Authentication:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/5e550938-91d4-459f-b67d-75d70009e3f3
    """
    b_user = (username.upper() + (domain_name or "")).encode("utf-16-le")
    return hmac_md5(nt_hash, b_user)


def rc4(h: RC4Handle, d: bytes) -> bytes:
    """RC4 encryption of the data specified using the handle generated by rc4init."""
    return h.update(d)


def rc4k(k: bytes, d: bytes) -> bytes:
    """RC4 encryption with an explicit key.

    Indicates the encryption of data item `d` with the key `k` using the `RC4`_ algorithm.

    Args:
        k: The key to use for the RC4 cipher.
        d: The data to encrypt.

    Returns:
        bytes: The RC4 encrypted bytes.

    .. _RC4K:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/26c42637-9549-46ae-be2e-90f6f1360193
    """
    return rc4init(k).update(d)


def rc4init(k: bytes) -> RC4Handle:
    """Initialization of the RC4 handle using the key specified."""
    return RC4Handle(k)


def sealkey(flags: int, session_key: bytes, usage: str) -> bytes:
    """NTLM SEALKEY function.

    The MS-NLMP `SEALKEY`_ function used to generate the sealing keys for a security context.

    The pseudo-code for this function as documented under `SEALKEY`_ is::

        Define SEALKEY(NegFlg, ExportedSessionKey, Mode) as

            If (NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY flag is set in NegFlg)

                If ( NTLMSSP_NEGOTIATE_128 is set in NegFlg)
                    Set SealKey to ExportedSessionKey

                ElseIf ( NTLMSSP_NEGOTIATE_56 flag is set in NegFlg)
                    Set SealKey to ExportedSessionKey[0..6]

                Else
                    Set SealKey to ExportedSessionKey[0..4]

                Endif

                If (Mode equals "Client")
                    Set SealKey to MD5(ConcatenationOf(SealKey,
                        "session key to client-to-server sealing key magic constant"))

                Else
                    Set SealKey to MD5(ConcatenationOf(SealKey,
                        "session key to server-to-client sealing key magic constant"))

                Endif
            ElseIf ((NTLMSSP_NEGOTIATE_LM_KEY is set in NegFlg) or ((NTLMSSP_NEGOTIATE_DATAGRAM is set in NegFlg) and
                                                                    (NTLMRevisionCurrent >= NTLMSSP_REVISION_W2K3)))

                If (NTLMSSP_NEGOTIATE_56 flag is set in NegFlg)
                    Set SealKey to ConcatenationOf(ExportedSessionKey[0..6], 0xA0)

                Else
                    Set SealKey to ConcatenationOf(ExportedSessionKey[0..4], 0xE5, 0x38, 0xB0)

                EndIf

            Else
                Set SealKey to ExportedSessionKey
            Endif
        EndDefine

    Args:
        flags: The negotiated flags between the initiator and acceptor.
        session_key: The derived session key.
        usage: Whether the sealing key is for the 'initiate' or 'accept' context.

    Returns:
        bytes: The derived sealing key.

    .. _SEALKEY:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/bf39181d-e95d-40d7-a740-ab4ec3dc363d
    """
    if flags & NegotiateFlags.extended_session_security:
        if flags & NegotiateFlags.key_128:
            seal_key = session_key

        elif flags & NegotiateFlags.key_56:
            seal_key = session_key[:7]

        else:
            seal_key = session_key[:5]

        direction = b"client-to-server" if usage == "initiate" else b"server-to-client"

        return md5(seal_key + b"session key to %s sealing key magic constant\x00" % direction)

    elif flags & NegotiateFlags.lm_key or flags & NegotiateFlags.datagram:
        if flags & NegotiateFlags.key_56:
            return session_key[:7] + b"\xa0"

        else:
            return session_key[:5] + b"\xe5\x38\xb0"

    else:
        return session_key


def signkey(flags: int, session_key: bytes, usage: str) -> bytes:
    """NTLM SIGNKEY function.

    The MS-NLMP `SIGNKEY`_ function used to generate the signing keys for a security context.

    The pseudo-code for this function as documented under `SIGNKEY`_ is::

        Define SIGNKEY(NegFlg, ExportedSessionKey, Mode) as

            If (NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY flag is set in NegFlg)
                If (Mode equals "Client")
                    Set SignKey to MD5(ConcatenationOf(ExportedSessionKey,
                        "session key to client-to-server signing key magic constant"))

                Else
                    Set SignKey to MD5(ConcatenationOf(ExportedSessionKey,
                        "session key to server-to-client signing key magic constant"))

                Endif
            Else
                Set  SignKey to NIL

            Endif
        EndDefine

    Args:
        flags: The negotiated flags between the initiator and acceptor.
        session_key: The derived session key.
        usage: Whether the signing key is for the 'initiate' or 'accept' context.

    Returns:
        bytes: The derived singing key.

    .. _SIGNKEY:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/524cdccb-563e-4793-92b0-7bc321fce096
    """
    if flags & NegotiateFlags.extended_session_security == 0:
        return b""

    direction = b"client-to-server" if usage == "initiate" else b"server-to-client"

    return md5(session_key + b"session key to %s signing key magic constant\x00" % direction)



# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import enum
import struct
import typing

from spnego._asn1 import (
    TagClass,
    TypeTagNumber,
    get_sequence_value,
    pack_asn1,
    pack_asn1_bit_string,
    pack_asn1_enumerated,
    pack_asn1_general_string,
    pack_asn1_object_identifier,
    pack_asn1_octet_string,
    pack_asn1_sequence,
    unpack_asn1,
    unpack_asn1_bit_string,
    unpack_asn1_enumerated,
    unpack_asn1_general_string,
    unpack_asn1_object_identifier,
    unpack_asn1_octet_string,
    unpack_asn1_sequence,
    unpack_asn1_tagged_sequence,
)
from spnego._context import GSSMech
from spnego._kerberos import KerberosV5Msg
from spnego._ntlm_raw.messages import NTLMMessage


def pack_mech_type_list(
    mech_list: typing.Union[str, typing.List[str], typing.Tuple[str, ...], typing.Set[str]],
) -> bytes:
    """Packs a list of OIDs for the mechListMIC value.

    Will pack a list of object identifiers to the raw byte string value for the mechListMIC.

    Args:
        mech_list: The list of OIDs to back

    Returns:
        bytes: The byte string of the packed ASN.1 MechTypeList SEQUENCE OF value.
    """
    if not isinstance(mech_list, (list, tuple, set)):
        mech_list = [mech_list]

    return pack_asn1_sequence([pack_asn1_object_identifier(oid) for oid in mech_list])


def unpack_token(
    b_data: bytes,
    mech: typing.Optional[GSSMech] = None,
    unwrap: bool = False,
    encoding: typing.Optional[str] = None,
) -> typing.Any:
    """Unpacks a raw GSSAPI/SPNEGO token to a Python object.

    Unpacks the byte string into a Python object that represents the token passed in. This can return many different
    token types such as:

    * NTLM message(s)
    * SPNEGO/Negotiate init or response
    * Kerberos message(s)

    Args:
        b_data: The raw byte string to unpack.
        mech: A hint as to what the byte string is for.
        unwrap: Whether to unwrap raw bytes to a structured message or return the raw tokens bytes.
        encoding: Optional encoding used when unwrapping NTLM messages.

    Returns:
        any: The unpacked SPNEGO, Kerberos, or NTLM token.
    """
    # First check if the message is an NTLM message.
    if b_data.startswith(b"NTLMSSP\x00"):
        if unwrap:
            return NTLMMessage.unpack(b_data, encoding=encoding)

        else:
            return b_data

    if mech and mech.is_kerberos_oid:
        # A Kerberos value inside an InitialContextToken contains 2 bytes which we ignore.
        raw_data = unpack_asn1(b_data[2:])[0]

    else:
        raw_data = unpack_asn1(b_data)[0]

    if raw_data.tag_class == TagClass.application and mech and mech.is_kerberos_oid:
        return KerberosV5Msg.unpack(unpack_asn1(raw_data.b_data)[0])

    elif raw_data.tag_class == TagClass.application:
        # The first token is encapsulated in an InitialContextToken.
        if raw_data.tag_number != 0:
            raise ValueError("Expecting a tag number of 0 not %s for InitialContextToken" % raw_data.tag_number)

        initial_context_token = InitialContextToken.unpack(raw_data.b_data)

        # unwrap=True is called from python -m spnego and we don't want to loose any info in the output.
        if unwrap:
            return initial_context_token

        this_mech: typing.Optional[GSSMech]
        try:
            this_mech = GSSMech.from_oid(initial_context_token.this_mech)
        except ValueError:
            this_mech = None

        # We currently only support SPNEGO, or raw Kerberos here.
        if this_mech and (this_mech == GSSMech.spnego or (this_mech.is_kerberos_oid and unwrap)):
            return unpack_token(initial_context_token.inner_context_token, mech=this_mech)

        return b_data

    elif raw_data.tag_class == TagClass.context_specific:
        # This is a raw NegotiationToken that is wrapped in a CHOICE or 0 or 1.
        if raw_data.tag_number == 0:
            return NegTokenInit.unpack(raw_data.b_data)

        elif raw_data.tag_number == 1:
            return NegTokenResp.unpack(raw_data.b_data)

        else:
            raise ValueError("Unknown NegotiationToken CHOICE %d, only expecting 0 or 1" % raw_data.tag_number)

    elif unwrap:
        # Could also be the ASN.1 Sequence of the Kerberos message.
        return KerberosV5Msg.unpack(raw_data)

    else:
        return b_data


# https://www.rfc-editor.org/rfc/rfc4178.html#section-4.2.1 - ContextFlags
class ContextFlags(enum.IntFlag):
    deleg = 0
    mutual = 1
    replay = 2
    sequence = 3
    anon = 4
    conf = 5
    integ = 6

    @classmethod
    def native_labels(cls) -> typing.Dict["ContextFlags", str]:
        return {
            ContextFlags.deleg: "delegFlag",
            ContextFlags.mutual: "mutualFlag",
            ContextFlags.replay: "replayFlag",
            ContextFlags.sequence: "sequenceFlag",
            ContextFlags.anon: "anonFlag",
            ContextFlags.conf: "confFlag",
            ContextFlags.integ: "integFlag",
        }


# https://www.rfc-editor.org/rfc/rfc4178.html#section-4.2.2 - negState
class NegState(enum.IntEnum):
    accept_complete = 0
    accept_incomplete = 1
    reject = 2
    request_mic = 3

    @classmethod
    def native_labels(cls) -> typing.Dict["NegState", str]:
        return {
            NegState.accept_complete: "accept-complete",
            NegState.accept_incomplete: "accept-incomplete",
            NegState.reject: "reject",
            NegState.request_mic: "request-mic",
        }


class InitialContextToken:
    """GSSAPI InitialContextToken object.

    The InitialContextToken is the ASN.1 structure that contains the first GSSAPI token that is sent across the wire.
    The ASN.1 definition for this structure is defined in `RFC 2743 3.1`_::

        MechType ::= OBJECT IDENTIFIER
        -- data structure definitions
        -- callers must be able to distinguish among
        -- InitialContextToken, SubsequentContextToken,
        -- PerMsgToken, and SealedMessage data elements
        -- based on the usage in which they occur

        InitialContextToken ::=
        -- option indication (delegation, etc.) indicated within
        -- mechanism-specific token
        [APPLICATION 0] IMPLICIT SEQUENCE {
            thisMech MechType,
            innerContextToken ANY DEFINED BY thisMech
               -- contents mechanism-specific
               -- ASN.1 structure not required
            }

    Args:
        mech: The OID that defines the structure of the `token`.
        token: The token of the GSSAPI value.

    Attributes:
        this_mech (str): The object identifier that identifies what the inner_context_token is for.
        inner_context_token (bytes): The token value as defined by `this_mech`.

    .. _RFC 2743 3.1:
        https://www.rfc-editor.org/rfc/rfc2743#section-3.1.
    """

    def __init__(self, mech: typing.Union[GSSMech, str], token: bytes) -> None:
        if isinstance(mech, GSSMech):
            mech = mech.value

        self.this_mech = mech
        self.inner_context_token = token

    @property
    def token(self) -> typing.Any:
        mech: typing.Optional[GSSMech]
        try:
            mech = GSSMech.from_oid(self.this_mech)
        except ValueError:
            mech = None

        return unpack_token(self.inner_context_token, mech=mech)

    def pack(self) -> bytes:
        """Packs the InitialContextToken as a byte string."""
        return pack_asn1(
            TagClass.application,
            True,
            0,
            pack_asn1_object_identifier(self.this_mech, tag=True) + self.inner_context_token,
        )

    @staticmethod
    def unpack(b_data: bytes) -> "InitialContextToken":
        """Unpacks the InitialContextToken TLV value."""
        this_mech, inner_context_token = unpack_asn1(b_data)
        mech = unpack_asn1_object_identifier(this_mech)

        return InitialContextToken(mech, inner_context_token)


class NegTokenInit:
    """The NegTokenInit GSSAPI value.

    This is the initial negotiation message token in a GSSAPI exchange. Typically the `NegTokenInit` value is sent
    when sending the first authentication token. The `NegTokenInit2` token is an extension that adds the `negHints`
    field. Unfortunately as the tag number for the `mechListMIC` is the same for `negHints` unpacking the value
    requires some extra checks.

    The ASN.1 definition for the NegTokenInit structure is defined in `RFC 4178 4.2.1`_::

        NegTokenInit ::= SEQUENCE {
            mechTypes       [0] MechTypeList,
            reqFlags        [1] ContextFlags  OPTIONAL,
            -- inherited from RFC 2478 for backward compatibility,
            -- RECOMMENDED to be left out
            mechToken       [2] OCTET STRING  OPTIONAL,
            mechListMIC     [3] OCTET STRING  OPTIONAL,
            ...
        }
        ContextFlags ::= BIT STRING {
            delegFlag       (0),
            mutualFlag      (1),
            replayFlag      (2),
            sequenceFlag    (3),
            anonFlag        (4),
            confFlag        (5),
            integFlag       (6)
        } (SIZE (32))

    The ASN.1 definition for the `NegTokenInit2`_ structure is defined as::

        NegHints ::= SEQUENCE {
            hintName[0] GeneralString OPTIONAL,
            hintAddress[1] OCTET STRING OPTIONAL
        }
        NegTokenInit2 ::= SEQUENCE {
            mechTypes[0] MechTypeList OPTIONAL,
            reqFlags [1] ContextFlags OPTIONAL,
            mechToken [2] OCTET STRING OPTIONAL,
            negHints [3] NegHints OPTIONAL,
            mechListMIC [4] OCTET STRING OPTIONAL,
            ...
        }

    Args:
        mech_types: One or more security mechanisms available for the initiator, in decreasing preference order.
        req_flags: Should be omitted, service options that are requested to establish the context.
        mech_token: Contains the optimistic mechanism token.
        hint_name: Used for the NegTokenInit2 structure only, should be omitted.
        hint_address: Used for the NegTokenINit2 structure only, should be omitted.
        mech_list_mic: The message integrity code (MIC) token.

    Attributes:
        mech_types (List[str]): See args.
        req_flags (ContextFlags): See args.
        mech_token (bytes): See args.
        hint_name (bytes): See args.
        hint_address (bytes): See args.
        mech_list_mic (bytes): See args.

    .. _RFC 4178 4.2.1:
        https://www.rfc-editor.org/rfc/rfc4178.html#section-4.2.1

    .. _NegTokenInit2:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-spng/8e71cf53-e867-4b79-b5b5-38c92be3d472
    """

    def __init__(
        self,
        mech_types: typing.Optional[typing.List[str]] = None,
        req_flags: typing.Optional[ContextFlags] = None,
        mech_token: typing.Optional[bytes] = None,
        hint_name: typing.Optional[bytes] = None,
        hint_address: typing.Optional[bytes] = None,
        mech_list_mic: typing.Optional[bytes] = None,
    ) -> None:
        self.mech_types = mech_types or []
        self.req_flags = req_flags
        self.mech_token = mech_token
        self.hint_name = hint_name
        self.hint_address = hint_address
        self.mech_list_mic = mech_list_mic

    def pack(self) -> bytes:
        """Packs the NegTokenInit as a byte string."""

        def pack_elements(
            value_map: typing.Iterable[typing.Tuple[int, typing.Any, typing.Callable]],
        ) -> typing.List[bytes]:
            elements = []
            for tag, value, pack_func in value_map:
                if value is not None:
                    elements.append(pack_asn1(TagClass.context_specific, True, tag, pack_func(value)))

            return elements

        req_flags = struct.pack("B", self.req_flags) if self.req_flags is not None else None
        base_map: typing.List[typing.Tuple[int, typing.Any, typing.Callable]] = [
            (0, self.mech_types, pack_mech_type_list),
            (1, req_flags, pack_asn1_bit_string),
            (2, self.mech_token, pack_asn1_octet_string),
        ]

        # The placement of the mechListMIC is dependent on whether we are packing a NegTokenInit with or without the
        # negHints field.
        neg_hints = pack_elements(
            [
                (0, self.hint_name, pack_asn1_general_string),
                (1, self.hint_address, pack_asn1_octet_string),
            ]
        )

        if neg_hints:
            base_map.append((3, neg_hints, pack_asn1_sequence))
            base_map.append((4, self.mech_list_mic, pack_asn1_octet_string))

        else:
            base_map.append((3, self.mech_list_mic, pack_asn1_octet_string))

        init_sequence = pack_elements(base_map)

        # The NegTokenInit will always be wrapped in an InitialContextToken -> NegotiationToken - CHOICE 0.
        b_data = pack_asn1_sequence(init_sequence)
        return InitialContextToken(GSSMech.spnego.value, pack_asn1(TagClass.context_specific, True, 0, b_data)).pack()

    @staticmethod
    def unpack(b_data: bytes) -> "NegTokenInit":
        """Unpacks the NegTokenInit TLV value."""
        neg_seq = unpack_asn1_tagged_sequence(unpack_asn1(b_data)[0])

        mech_types = [
            unpack_asn1_object_identifier(m)
            for m in get_sequence_value(neg_seq, 0, "NegTokenInit", "mechTypes", unpack_asn1_sequence) or []
        ]

        req_flags = get_sequence_value(neg_seq, 1, "NegTokenInit", "reqFlags", unpack_asn1_bit_string)
        if req_flags:
            # Can be up to 32 bits in length but RFC 4178 states "Implementations should not expect to receive exactly
            # 32 bits in an encoding of ContextFlags." The spec also documents req flags up to 6 so let's just get the
            # last byte. In reality we shouldn't ever receive this but it's left here for posterity.
            req_flags = ContextFlags(bytearray(req_flags)[-1])

        mech_token = get_sequence_value(neg_seq, 2, "NegTokenInit", "mechToken", unpack_asn1_octet_string)

        hint_name = hint_address = mech_list_mic = None
        if 3 in neg_seq:
            # Microsoft helpfully sends a NegTokenInit2 payload which sets 'negHints [3] NegHints OPTIONAL' and the
            # mechListMIC is actually at the 4th sequence entry. Because the NegTokenInit2 has the same choice in
            # NegotiationToken as NegTokenInit ([0]) we can only differentiate when unpacking based on the class/tags.
            tag_class = neg_seq[3].tag_class
            tag_number = neg_seq[3].tag_number

            if tag_class == TagClass.universal and tag_number == TypeTagNumber.sequence:
                neg_hints = unpack_asn1_tagged_sequence(neg_seq[3].b_data)

                # Windows 2000, 2003, and XP put the SPN encoded as the OEM code page, because there's no sane way
                # to decode this without prior knowledge a GeneralString stays a byte string in Python.
                # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-spng/211417c4-11ef-46c0-a8fb-f178a51c2088#Appendix_A_5
                hint_name = get_sequence_value(neg_hints, 0, "NegHints", "hintName", unpack_asn1_general_string)
                hint_address = get_sequence_value(neg_hints, 1, "NegHints", "hintAddress", unpack_asn1_octet_string)

            else:
                # Wasn't a sequence, should be mechListMIC.
                mech_list_mic = get_sequence_value(neg_seq, 3, "NegTokenInit", "mechListMIC", unpack_asn1_octet_string)

        if not mech_list_mic:
            mech_list_mic = get_sequence_value(neg_seq, 4, "NegTokenInit2", "mechListMIC", unpack_asn1_octet_string)

        return NegTokenInit(mech_types, req_flags, mech_token, hint_name, hint_address, mech_list_mic)


class NegTokenResp:
    """The NegTokenResp GSSAPI value.

    This is the message token in a GSSAPI exchange that is used for subsequent messages after the `NegTokenInit` has
    been exchanged.

    The ASN.1 definition for the NegTokenResp structure is defined in `RFC 4178 4.2.2`_::

        NegTokenResp ::= SEQUENCE {
            negState       [0] ENUMERATED {
                accept-completed    (0),
                accept-incomplete   (1),
                reject              (2),
                request-mic         (3)
            }                                 OPTIONAL,
            -- REQUIRED in the first reply from the target
            supportedMech   [1] MechType      OPTIONAL,
            -- present only in the first reply from the target
            responseToken   [2] OCTET STRING  OPTIONAL,
            mechListMIC     [3] OCTET STRING  OPTIONAL,
            ...
        }

    Args:
        neg_state: The state of the negotiation.
        supported_mech: Should only be present in the first reply, must be one of the mech(s) offered by the initiator.
        response_token: Contains the token specific to the mechanism selected.
        mech_list_mic: The message integrity code (MIC) token.

    Attributes:
        neg_state (NegState): See args.
        supported_mech (str): See args.
        response_token (bytes): See args.
        mech_list_mic (bytes): See args.

    .. _RFC 4178 4.2.2:
        https://www.rfc-editor.org/rfc/rfc4178.html#section-4.2.2
    """

    def __init__(
        self,
        neg_state: typing.Optional[NegState] = None,
        supported_mech: typing.Optional[str] = None,
        response_token: typing.Optional[bytes] = None,
        mech_list_mic: typing.Optional[bytes] = None,
    ) -> None:
        self.neg_state = neg_state
        self.supported_mech = supported_mech
        self.response_token = response_token
        self.mech_list_mic = mech_list_mic

    def pack(self) -> bytes:
        """Packs the NegTokenResp as a byte string."""
        value_map: typing.List[typing.Tuple[int, typing.Any, typing.Callable[[typing.Any], bytes]]] = [
            (0, self.neg_state, pack_asn1_enumerated),
            (1, self.supported_mech, pack_asn1_object_identifier),
            (2, self.response_token, pack_asn1_octet_string),
            (3, self.mech_list_mic, pack_asn1_octet_string),
        ]
        elements = []
        for tag, value, pack_func in value_map:
            if value is not None:
                elements.append(pack_asn1(TagClass.context_specific, True, tag, pack_func(value)))

        # The NegTokenResp will always be wrapped NegotiationToken - CHOICE 1.
        b_data = pack_asn1_sequence(elements)
        return pack_asn1(TagClass.context_specific, True, 1, b_data)

    @staticmethod
    def unpack(b_data: bytes) -> "NegTokenResp":
        """Unpacks the NegTokenResp TLV value."""
        neg_seq = unpack_asn1_tagged_sequence(unpack_asn1(b_data)[0])

        neg_state = get_sequence_value(neg_seq, 0, "NegTokenResp", "negState", unpack_asn1_enumerated)
        if neg_state is not None:
            neg_state = NegState(neg_state)

        supported_mech = get_sequence_value(neg_seq, 1, "NegTokenResp", "supportedMech", unpack_asn1_object_identifier)
        response_token = get_sequence_value(neg_seq, 2, "NegTokenResp", "responseToken", unpack_asn1_octet_string)
        mech_list_mic = get_sequence_value(neg_seq, 3, "NegTokenResp", "mechListMIC", unpack_asn1_octet_string)

        return NegTokenResp(neg_state, supported_mech, response_token, mech_list_mic)



# Copyright: (c) 2022, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import enum
import typing


def _add_missing_enum_member(
    cls: typing.Type[enum.IntEnum],
    value: object,
    label: str,
) -> typing.Optional[enum.Enum]:
    if not isinstance(value, int):
        return None

    new_member = int.__new__(cls)
    new_member._name_ = label.format(value)
    new_member._value_ = value
    return cls._value2member_map_.setdefault(value, new_member)


class TlsProtocolVersion(enum.IntEnum):
    tls1_0 = 0x0301
    tls1_1 = 0x0302
    tls1_2 = 0x0303
    tls1_3 = 0x0304

    @classmethod
    def native_labels(cls) -> typing.Dict["TlsProtocolVersion", str]:
        return {
            TlsProtocolVersion.tls1_0: "TLS 1.0 (0x0301)",
            TlsProtocolVersion.tls1_1: "TLS 1.1 (0x0302)",
            TlsProtocolVersion.tls1_2: "TLS 1.2 (0x0303)",
            TlsProtocolVersion.tls1_3: "TLS 1.3 (0x0304)",
        }

    @classmethod
    def _missing_(cls, value: object) -> typing.Optional[enum.Enum]:
        return _add_missing_enum_member(cls, value, "Unknown TLS Protocol Version 0x{0:04X}")


class TlsContentType(enum.IntEnum):
    invalid = 0
    change_cipher_spec = 20
    alert = 21
    handshake = 22
    application_data = 23

    @classmethod
    def _missing_(cls, value: object) -> typing.Optional[enum.Enum]:
        return _add_missing_enum_member(cls, value, "Unknown TLS Content Type 0x{0:02X}")


class TlsHandshakeMessageType(enum.IntEnum):
    hello_request = 0
    client_hello = 1
    server_hello = 2
    hello_verify_request = 3
    new_session_ticket = 4
    end_of_early_data = 5
    hello_retry_request = 6
    encrypted_extensions = 8
    request_connection_id = 9
    new_connection_id = 10
    certificate = 11
    server_key_exchange = 12
    certificate_request = 13
    server_hello_done = 14
    certificate_verify = 15
    client_key_exchange = 16
    client_certificate_request = 17
    finished = 20
    certificate_url = 21
    certificate_status = 22
    supplemental_data = 23
    key_update = 24
    compressed_certificate = 25
    ekt_key = 26
    message_hash = 254

    @classmethod
    def _missing_(cls, value: object) -> typing.Optional[enum.Enum]:
        return _add_missing_enum_member(cls, value, "Unknown Handshake Message Type 0x{0:02X}")


class TlsCipherSuite(enum.IntEnum):
    TLS_NULL_WITH_NULL_NULL = 0x0000
    TLS_RSA_WITH_NULL_MD5 = 0x0001
    TLS_RSA_WITH_NULL_SHA = 0x0002
    TLS_RSA_EXPORT_WITH_RC4_40_MD5 = 0x0003
    TLS_RSA_WITH_RC4_128_MD5 = 0x0004
    TLS_RSA_WITH_RC4_128_SHA = 0x0005
    TLS_RSA_EXPORT_WITH_RC2_CBC_40_MD5 = 0x0006
    TLS_RSA_WITH_IDEA_CBC_SHA = 0x0007
    TLS_RSA_EXPORT_WITH_DES40_CBC_SHA = 0x0008
    TLS_RSA_WITH_DES_CBC_SHA = 0x0009
    TLS_RSA_WITH_3DES_EDE_CBC_SHA = 0x000A
    TLS_DH_DSS_EXPORT_WITH_DES40_CBC_SHA = 0x000B
    TLS_DH_DSS_WITH_DES_CBC_SHA = 0x000C
    TLS_DH_DSS_WITH_3DES_EDE_CBC_SHA = 0x000D
    TLS_DH_RSA_EXPORT_WITH_DES40_CBC_SHA = 0x000E
    TLS_DH_RSA_WITH_DES_CBC_SHA = 0x000F
    TLS_DH_RSA_WITH_3DES_EDE_CBC_SHA = 0x0010
    TLS_DHE_DSS_EXPORT_WITH_DES40_CBC_SHA = 0x0011
    TLS_DHE_DSS_WITH_DES_CBC_SHA = 0x0012
    TLS_DHE_DSS_WITH_3DES_EDE_CBC_SHA = 0x0013
    TLS_DHE_RSA_EXPORT_WITH_DES40_CBC_SHA = 0x0014
    TLS_DHE_RSA_WITH_DES_CBC_SHA = 0x0015
    TLS_DHE_RSA_WITH_3DES_EDE_CBC_SHA = 0x0016
    TLS_DH_anon_EXPORT_WITH_RC4_40_MD5 = 0x0017
    TLS_DH_anon_WITH_RC4_128_MD5 = 0x0018
    TLS_DH_anon_EXPORT_WITH_DES40_CBC_SHA = 0x0019
    TLS_DH_anon_WITH_DES_CBC_SHA = 0x001A
    TLS_DH_anon_WITH_3DES_EDE_CBC_SHA = 0x001B
    TLS_KRB5_WITH_DES_CBC_SHA = 0x001E
    TLS_KRB5_WITH_3DES_EDE_CBC_SHA = 0x001F
    TLS_KRB5_WITH_RC4_128_SHA = 0x0020
    TLS_KRB5_WITH_IDEA_CBC_SHA = 0x0021
    TLS_KRB5_WITH_DES_CBC_MD5 = 0x0022
    TLS_KRB5_WITH_3DES_EDE_CBC_MD5 = 0x0023
    TLS_KRB5_WITH_RC4_128_MD5 = 0x0024
    TLS_KRB5_WITH_IDEA_CBC_MD5 = 0x0025
    TLS_KRB5_EXPORT_WITH_DES_CBC_40_SHA = 0x0026
    TLS_KRB5_EXPORT_WITH_RC2_CBC_40_SHA = 0x0027
    TLS_KRB5_EXPORT_WITH_RC4_40_SHA = 0x0028
    TLS_KRB5_EXPORT_WITH_DES_CBC_40_MD5 = 0x0029
    TLS_KRB5_EXPORT_WITH_RC2_CBC_40_MD5 = 0x002A
    TLS_KRB5_EXPORT_WITH_RC4_40_MD5 = 0x002B
    TLS_PSK_WITH_NULL_SHA = 0x002C
    TLS_DHE_PSK_WITH_NULL_SHA = 0x002D
    TLS_RSA_PSK_WITH_NULL_SHA = 0x002E
    TLS_RSA_WITH_AES_128_CBC_SHA = 0x002F
    TLS_DH_DSS_WITH_AES_128_CBC_SHA = 0x0030
    TLS_DH_RSA_WITH_AES_128_CBC_SHA = 0x0031
    TLS_DHE_DSS_WITH_AES_128_CBC_SHA = 0x0032
    TLS_DHE_RSA_WITH_AES_128_CBC_SHA = 0x0033
    TLS_DH_anon_WITH_AES_128_CBC_SHA = 0x0034
    TLS_RSA_WITH_AES_256_CBC_SHA = 0x0035
    TLS_DH_DSS_WITH_AES_256_CBC_SHA = 0x0036
    TLS_DH_RSA_WITH_AES_256_CBC_SHA = 0x0037
    TLS_DHE_DSS_WITH_AES_256_CBC_SHA = 0x0038
    TLS_DHE_RSA_WITH_AES_256_CBC_SHA = 0x0039
    TLS_DH_anon_WITH_AES_256_CBC_SHA = 0x003A
    TLS_RSA_WITH_NULL_SHA256 = 0x003B
    TLS_RSA_WITH_AES_128_CBC_SHA256 = 0x003C
    TLS_RSA_WITH_AES_256_CBC_SHA256 = 0x003D
    TLS_DH_DSS_WITH_AES_128_CBC_SHA256 = 0x003E
    TLS_DH_RSA_WITH_AES_128_CBC_SHA256 = 0x003F
    TLS_DHE_DSS_WITH_AES_128_CBC_SHA256 = 0x0040
    TLS_RSA_WITH_CAMELLIA_128_CBC_SHA = 0x0041
    TLS_DH_DSS_WITH_CAMELLIA_128_CBC_SHA = 0x0042
    TLS_DH_RSA_WITH_CAMELLIA_128_CBC_SHA = 0x0043
    TLS_DHE_DSS_WITH_CAMELLIA_128_CBC_SHA = 0x0044
    TLS_DHE_RSA_WITH_CAMELLIA_128_CBC_SHA = 0x0045
    TLS_DH_anon_WITH_CAMELLIA_128_CBC_SHA = 0x0046
    TLS_DHE_RSA_WITH_AES_128_CBC_SHA256 = 0x0067
    TLS_DH_DSS_WITH_AES_256_CBC_SHA256 = 0x0068
    TLS_DH_RSA_WITH_AES_256_CBC_SHA256 = 0x0069
    TLS_DHE_DSS_WITH_AES_256_CBC_SHA256 = 0x006A
    TLS_DHE_RSA_WITH_AES_256_CBC_SHA256 = 0x006B
    TLS_DH_anon_WITH_AES_128_CBC_SHA256 = 0x006C
    TLS_DH_anon_WITH_AES_256_CBC_SHA256 = 0x006D
    TLS_RSA_WITH_CAMELLIA_256_CBC_SHA = 0x0084
    TLS_DH_DSS_WITH_CAMELLIA_256_CBC_SHA = 0x0085
    TLS_DH_RSA_WITH_CAMELLIA_256_CBC_SHA = 0x0086
    TLS_DHE_DSS_WITH_CAMELLIA_256_CBC_SHA = 0x0087
    TLS_DHE_RSA_WITH_CAMELLIA_256_CBC_SHA = 0x0088
    TLS_DH_anon_WITH_CAMELLIA_256_CBC_SHA = 0x0089
    TLS_PSK_WITH_RC4_128_SHA = 0x008A
    TLS_PSK_WITH_3DES_EDE_CBC_SHA = 0x008B
    TLS_PSK_WITH_AES_128_CBC_SHA = 0x008C
    TLS_PSK_WITH_AES_256_CBC_SHA = 0x008D
    TLS_DHE_PSK_WITH_RC4_128_SHA = 0x008E
    TLS_DHE_PSK_WITH_3DES_EDE_CBC_SHA = 0x008F
    TLS_DHE_PSK_WITH_AES_128_CBC_SHA = 0x0090
    TLS_DHE_PSK_WITH_AES_256_CBC_SHA = 0x0091
    TLS_RSA_PSK_WITH_RC4_128_SHA = 0x0092
    TLS_RSA_PSK_WITH_3DES_EDE_CBC_SHA = 0x0093
    TLS_RSA_PSK_WITH_AES_128_CBC_SHA = 0x0094
    TLS_RSA_PSK_WITH_AES_256_CBC_SHA = 0x0095
    TLS_RSA_WITH_SEED_CBC_SHA = 0x0096
    TLS_DH_DSS_WITH_SEED_CBC_SHA = 0x0097
    TLS_DH_RSA_WITH_SEED_CBC_SHA = 0x0098
    TLS_DHE_DSS_WITH_SEED_CBC_SHA = 0x0099
    TLS_DHE_RSA_WITH_SEED_CBC_SHA = 0x009A
    TLS_DH_anon_WITH_SEED_CBC_SHA = 0x009B
    TLS_RSA_WITH_AES_128_GCM_SHA256 = 0x009C
    TLS_RSA_WITH_AES_256_GCM_SHA384 = 0x009D
    TLS_DHE_RSA_WITH_AES_128_GCM_SHA256 = 0x009E
    TLS_DHE_RSA_WITH_AES_256_GCM_SHA384 = 0x009F
    TLS_DH_RSA_WITH_AES_128_GCM_SHA256 = 0x00A0
    TLS_DH_RSA_WITH_AES_256_GCM_SHA384 = 0x00A1
    TLS_DHE_DSS_WITH_AES_128_GCM_SHA256 = 0x00A2
    TLS_DHE_DSS_WITH_AES_256_GCM_SHA384 = 0x00A3
    TLS_DH_DSS_WITH_AES_128_GCM_SHA256 = 0x00A4
    TLS_DH_DSS_WITH_AES_256_GCM_SHA384 = 0x00A5
    TLS_DH_anon_WITH_AES_128_GCM_SHA256 = 0x00A6
    TLS_DH_anon_WITH_AES_256_GCM_SHA384 = 0x00A7
    TLS_PSK_WITH_AES_128_GCM_SHA256 = 0x00A8
    TLS_PSK_WITH_AES_256_GCM_SHA384 = 0x00A9
    TLS_DHE_PSK_WITH_AES_128_GCM_SHA256 = 0x00AA
    TLS_DHE_PSK_WITH_AES_256_GCM_SHA384 = 0x00AB
    TLS_RSA_PSK_WITH_AES_128_GCM_SHA256 = 0x00AC
    TLS_RSA_PSK_WITH_AES_256_GCM_SHA384 = 0x00AD
    TLS_PSK_WITH_AES_128_CBC_SHA256 = 0x00AE
    TLS_PSK_WITH_AES_256_CBC_SHA384 = 0x00AF
    TLS_PSK_WITH_NULL_SHA256 = 0x00B0
    TLS_PSK_WITH_NULL_SHA384 = 0x00B1
    TLS_DHE_PSK_WITH_AES_128_CBC_SHA256 = 0x00B2
    TLS_DHE_PSK_WITH_AES_256_CBC_SHA384 = 0x00B3
    TLS_DHE_PSK_WITH_NULL_SHA256 = 0x00B4
    TLS_DHE_PSK_WITH_NULL_SHA384 = 0x00B5
    TLS_RSA_PSK_WITH_AES_128_CBC_SHA256 = 0x00B6
    TLS_RSA_PSK_WITH_AES_256_CBC_SHA384 = 0x00B7
    TLS_RSA_PSK_WITH_NULL_SHA256 = 0x00B8
    TLS_RSA_PSK_WITH_NULL_SHA384 = 0x00B9
    TLS_RSA_WITH_CAMELLIA_128_CBC_SHA256 = 0x00BA
    TLS_DH_DSS_WITH_CAMELLIA_128_CBC_SHA256 = 0x00BB
    TLS_DH_RSA_WITH_CAMELLIA_128_CBC_SHA256 = 0x00BC
    TLS_DHE_DSS_WITH_CAMELLIA_128_CBC_SHA256 = 0x00BD
    TLS_DHE_RSA_WITH_CAMELLIA_128_CBC_SHA256 = 0x00BE
    TLS_DH_anon_WITH_CAMELLIA_128_CBC_SHA256 = 0x00BF
    TLS_RSA_WITH_CAMELLIA_256_CBC_SHA256 = 0x00C0
    TLS_DH_DSS_WITH_CAMELLIA_256_CBC_SHA256 = 0x00C1
    TLS_DH_RSA_WITH_CAMELLIA_256_CBC_SHA256 = 0x00C2
    TLS_DHE_DSS_WITH_CAMELLIA_256_CBC_SHA256 = 0x00C3
    TLS_DHE_RSA_WITH_CAMELLIA_256_CBC_SHA256 = 0x00C4
    TLS_DH_anon_WITH_CAMELLIA_256_CBC_SHA256 = 0x00C5
    TLS_SM4_GCM_SM3 = 0x00C6
    TLS_SM4_CCM_SM3 = 0x00C7
    TLS_EMPTY_RENEGOTIATION_INFO_SCSV = 0x00FF
    TLS_AES_128_GCM_SHA256 = 0x1301
    TLS_AES_256_GCM_SHA384 = 0x1302
    TLS_CHACHA20_POLY1305_SHA256 = 0x1303
    TLS_AES_128_CCM_SHA256 = 0x1304
    TLS_AES_128_CCM_8_SHA256 = 0x1305
    TLS_FALLBACK_SCSV = 0x5600
    TLS_ECDH_ECDSA_WITH_NULL_SHA = 0xC001
    TLS_ECDH_ECDSA_WITH_RC4_128_SHA = 0xC002
    TLS_ECDH_ECDSA_WITH_3DES_EDE_CBC_SHA = 0xC003
    TLS_ECDH_ECDSA_WITH_AES_128_CBC_SHA = 0xC004
    TLS_ECDH_ECDSA_WITH_AES_256_CBC_SHA = 0xC005
    TLS_ECDHE_ECDSA_WITH_NULL_SHA = 0xC006
    TLS_ECDHE_ECDSA_WITH_RC4_128_SHA = 0xC007
    TLS_ECDHE_ECDSA_WITH_3DES_EDE_CBC_SHA = 0xC008
    TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA = 0xC009
    TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA = 0xC00A
    TLS_ECDH_RSA_WITH_NULL_SHA = 0xC00B
    TLS_ECDH_RSA_WITH_RC4_128_SHA = 0xC00C
    TLS_ECDH_RSA_WITH_3DES_EDE_CBC_SHA = 0xC00D
    TLS_ECDH_RSA_WITH_AES_128_CBC_SHA = 0xC00E
    TLS_ECDH_RSA_WITH_AES_256_CBC_SHA = 0xC00F
    TLS_ECDHE_RSA_WITH_NULL_SHA = 0xC010
    TLS_ECDHE_RSA_WITH_RC4_128_SHA = 0xC011
    TLS_ECDHE_RSA_WITH_3DES_EDE_CBC_SHA = 0xC012
    TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA = 0xC013
    TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA = 0xC014
    TLS_ECDH_anon_WITH_NULL_SHA = 0xC015
    TLS_ECDH_anon_WITH_RC4_128_SHA = 0xC016
    TLS_ECDH_anon_WITH_3DES_EDE_CBC_SHA = 0xC017
    TLS_ECDH_anon_WITH_AES_128_CBC_SHA = 0xC018
    TLS_ECDH_anon_WITH_AES_256_CBC_SHA = 0xC019
    TLS_SRP_SHA_WITH_3DES_EDE_CBC_SHA = 0xC01A
    TLS_SRP_SHA_RSA_WITH_3DES_EDE_CBC_SHA = 0xC01B
    TLS_SRP_SHA_DSS_WITH_3DES_EDE_CBC_SHA = 0xC01C
    TLS_SRP_SHA_WITH_AES_128_CBC_SHA = 0xC01D
    TLS_SRP_SHA_RSA_WITH_AES_128_CBC_SHA = 0xC01E
    TLS_SRP_SHA_DSS_WITH_AES_128_CBC_SHA = 0xC01F
    TLS_SRP_SHA_WITH_AES_256_CBC_SHA = 0xC020
    TLS_SRP_SHA_RSA_WITH_AES_256_CBC_SHA = 0xC021
    TLS_SRP_SHA_DSS_WITH_AES_256_CBC_SHA = 0xC022
    TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA256 = 0xC023
    TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA384 = 0xC024
    TLS_ECDH_ECDSA_WITH_AES_128_CBC_SHA256 = 0xC025
    TLS_ECDH_ECDSA_WITH_AES_256_CBC_SHA384 = 0xC026
    TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256 = 0xC027
    TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA384 = 0xC028
    TLS_ECDH_RSA_WITH_AES_128_CBC_SHA256 = 0xC029
    TLS_ECDH_RSA_WITH_AES_256_CBC_SHA384 = 0xC02A
    TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256 = 0xC02B
    TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384 = 0xC02C
    TLS_ECDH_ECDSA_WITH_AES_128_GCM_SHA256 = 0xC02D
    TLS_ECDH_ECDSA_WITH_AES_256_GCM_SHA384 = 0xC02E
    TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256 = 0xC02F
    TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384 = 0xC030
    TLS_ECDH_RSA_WITH_AES_128_GCM_SHA256 = 0xC031
    TLS_ECDH_RSA_WITH_AES_256_GCM_SHA384 = 0xC032
    TLS_ECDHE_PSK_WITH_RC4_128_SHA = 0xC033
    TLS_ECDHE_PSK_WITH_3DES_EDE_CBC_SHA = 0xC034
    TLS_ECDHE_PSK_WITH_AES_128_CBC_SHA = 0xC035
    TLS_ECDHE_PSK_WITH_AES_256_CBC_SHA = 0xC036
    TLS_ECDHE_PSK_WITH_AES_128_CBC_SHA256 = 0xC037
    TLS_ECDHE_PSK_WITH_AES_256_CBC_SHA384 = 0xC038
    TLS_ECDHE_PSK_WITH_NULL_SHA = 0xC039
    TLS_ECDHE_PSK_WITH_NULL_SHA256 = 0xC03A
    TLS_ECDHE_PSK_WITH_NULL_SHA384 = 0xC03B
    TLS_RSA_WITH_ARIA_128_CBC_SHA256 = 0xC03C
    TLS_RSA_WITH_ARIA_256_CBC_SHA384 = 0xC03D
    TLS_DH_DSS_WITH_ARIA_128_CBC_SHA256 = 0xC03E
    TLS_DH_DSS_WITH_ARIA_256_CBC_SHA384 = 0xC03F
    TLS_DH_RSA_WITH_ARIA_128_CBC_SHA256 = 0xC040
    TLS_DH_RSA_WITH_ARIA_256_CBC_SHA384 = 0xC041
    TLS_DHE_DSS_WITH_ARIA_128_CBC_SHA256 = 0xC042
    TLS_DHE_DSS_WITH_ARIA_256_CBC_SHA384 = 0xC043
    TLS_DHE_RSA_WITH_ARIA_128_CBC_SHA256 = 0xC044
    TLS_DHE_RSA_WITH_ARIA_256_CBC_SHA384 = 0xC045
    TLS_DH_anon_WITH_ARIA_128_CBC_SHA256 = 0xC046
    TLS_DH_anon_WITH_ARIA_256_CBC_SHA384 = 0xC047
    TLS_ECDHE_ECDSA_WITH_ARIA_128_CBC_SHA256 = 0xC048
    TLS_ECDHE_ECDSA_WITH_ARIA_256_CBC_SHA384 = 0xC049
    TLS_ECDH_ECDSA_WITH_ARIA_128_CBC_SHA256 = 0xC04A
    TLS_ECDH_ECDSA_WITH_ARIA_256_CBC_SHA384 = 0xC04B
    TLS_ECDHE_RSA_WITH_ARIA_128_CBC_SHA256 = 0xC04C
    TLS_ECDHE_RSA_WITH_ARIA_256_CBC_SHA384 = 0xC04D
    TLS_ECDH_RSA_WITH_ARIA_128_CBC_SHA256 = 0xC04E
    TLS_ECDH_RSA_WITH_ARIA_256_CBC_SHA384 = 0xC04F
    TLS_RSA_WITH_ARIA_128_GCM_SHA256 = 0xC050
    TLS_RSA_WITH_ARIA_256_GCM_SHA384 = 0xC051
    TLS_DHE_RSA_WITH_ARIA_128_GCM_SHA256 = 0xC052
    TLS_DHE_RSA_WITH_ARIA_256_GCM_SHA384 = 0xC053
    TLS_DH_RSA_WITH_ARIA_128_GCM_SHA256 = 0xC054
    TLS_DH_RSA_WITH_ARIA_256_GCM_SHA384 = 0xC055
    TLS_DHE_DSS_WITH_ARIA_128_GCM_SHA256 = 0xC056
    TLS_DHE_DSS_WITH_ARIA_256_GCM_SHA384 = 0xC057
    TLS_DH_DSS_WITH_ARIA_128_GCM_SHA256 = 0xC058
    TLS_DH_DSS_WITH_ARIA_256_GCM_SHA384 = 0xC059
    TLS_DH_anon_WITH_ARIA_128_GCM_SHA256 = 0xC05A
    TLS_DH_anon_WITH_ARIA_256_GCM_SHA384 = 0xC05B
    TLS_ECDHE_ECDSA_WITH_ARIA_128_GCM_SHA256 = 0xC05C
    TLS_ECDHE_ECDSA_WITH_ARIA_256_GCM_SHA384 = 0xC05D
    TLS_ECDH_ECDSA_WITH_ARIA_128_GCM_SHA256 = 0xC05E
    TLS_ECDH_ECDSA_WITH_ARIA_256_GCM_SHA384 = 0xC05F
    TLS_ECDHE_RSA_WITH_ARIA_128_GCM_SHA256 = 0xC060
    TLS_ECDHE_RSA_WITH_ARIA_256_GCM_SHA384 = 0xC061
    TLS_ECDH_RSA_WITH_ARIA_128_GCM_SHA256 = 0xC062
    TLS_ECDH_RSA_WITH_ARIA_256_GCM_SHA384 = 0xC063
    TLS_PSK_WITH_ARIA_128_CBC_SHA256 = 0xC064
    TLS_PSK_WITH_ARIA_256_CBC_SHA384 = 0xC065
    TLS_DHE_PSK_WITH_ARIA_128_CBC_SHA256 = 0xC066
    TLS_DHE_PSK_WITH_ARIA_256_CBC_SHA384 = 0xC067
    TLS_RSA_PSK_WITH_ARIA_128_CBC_SHA256 = 0xC068
    TLS_RSA_PSK_WITH_ARIA_256_CBC_SHA384 = 0xC069
    TLS_PSK_WITH_ARIA_128_GCM_SHA256 = 0xC06A
    TLS_PSK_WITH_ARIA_256_GCM_SHA384 = 0xC06B
    TLS_DHE_PSK_WITH_ARIA_128_GCM_SHA256 = 0xC06C
    TLS_DHE_PSK_WITH_ARIA_256_GCM_SHA384 = 0xC06D
    TLS_RSA_PSK_WITH_ARIA_128_GCM_SHA256 = 0xC06E
    TLS_RSA_PSK_WITH_ARIA_256_GCM_SHA384 = 0xC06F
    TLS_ECDHE_PSK_WITH_ARIA_128_CBC_SHA256 = 0xC070
    TLS_ECDHE_PSK_WITH_ARIA_256_CBC_SHA384 = 0xC071
    TLS_ECDHE_ECDSA_WITH_CAMELLIA_128_CBC_SHA256 = 0xC072
    TLS_ECDHE_ECDSA_WITH_CAMELLIA_256_CBC_SHA384 = 0xC073
    TLS_ECDH_ECDSA_WITH_CAMELLIA_128_CBC_SHA256 = 0xC074
    TLS_ECDH_ECDSA_WITH_CAMELLIA_256_CBC_SHA384 = 0xC075
    TLS_ECDHE_RSA_WITH_CAMELLIA_128_CBC_SHA256 = 0xC076
    TLS_ECDHE_RSA_WITH_CAMELLIA_256_CBC_SHA384 = 0xC077
    TLS_ECDH_RSA_WITH_CAMELLIA_128_CBC_SHA256 = 0xC078
    TLS_ECDH_RSA_WITH_CAMELLIA_256_CBC_SHA384 = 0xC079
    TLS_RSA_WITH_CAMELLIA_128_GCM_SHA256 = 0xC07A
    TLS_RSA_WITH_CAMELLIA_256_GCM_SHA384 = 0xC07B
    TLS_DHE_RSA_WITH_CAMELLIA_128_GCM_SHA256 = 0xC07C
    TLS_DHE_RSA_WITH_CAMELLIA_256_GCM_SHA384 = 0xC07D
    TLS_DH_RSA_WITH_CAMELLIA_128_GCM_SHA256 = 0xC07E
    TLS_DH_RSA_WITH_CAMELLIA_256_GCM_SHA384 = 0xC07F
    TLS_DHE_DSS_WITH_CAMELLIA_128_GCM_SHA256 = 0xC080
    TLS_DHE_DSS_WITH_CAMELLIA_256_GCM_SHA384 = 0xC081
    TLS_DH_DSS_WITH_CAMELLIA_128_GCM_SHA256 = 0xC082
    TLS_DH_DSS_WITH_CAMELLIA_256_GCM_SHA384 = 0xC083
    TLS_DH_anon_WITH_CAMELLIA_128_GCM_SHA256 = 0xC084
    TLS_DH_anon_WITH_CAMELLIA_256_GCM_SHA384 = 0xC085
    TLS_ECDHE_ECDSA_WITH_CAMELLIA_128_GCM_SHA256 = 0xC086
    TLS_ECDHE_ECDSA_WITH_CAMELLIA_256_GCM_SHA384 = 0xC087
    TLS_ECDH_ECDSA_WITH_CAMELLIA_128_GCM_SHA256 = 0xC088
    TLS_ECDH_ECDSA_WITH_CAMELLIA_256_GCM_SHA384 = 0xC089
    TLS_ECDHE_RSA_WITH_CAMELLIA_128_GCM_SHA256 = 0xC08A
    TLS_ECDHE_RSA_WITH_CAMELLIA_256_GCM_SHA384 = 0xC08B
    TLS_ECDH_RSA_WITH_CAMELLIA_128_GCM_SHA256 = 0xC08C
    TLS_ECDH_RSA_WITH_CAMELLIA_256_GCM_SHA384 = 0xC08D
    TLS_PSK_WITH_CAMELLIA_128_GCM_SHA256 = 0xC08E
    TLS_PSK_WITH_CAMELLIA_256_GCM_SHA384 = 0xC08F
    TLS_DHE_PSK_WITH_CAMELLIA_128_GCM_SHA256 = 0xC090
    TLS_DHE_PSK_WITH_CAMELLIA_256_GCM_SHA384 = 0xC091
    TLS_RSA_PSK_WITH_CAMELLIA_128_GCM_SHA256 = 0xC092
    TLS_RSA_PSK_WITH_CAMELLIA_256_GCM_SHA384 = 0xC093
    TLS_PSK_WITH_CAMELLIA_128_CBC_SHA256 = 0xC094
    TLS_PSK_WITH_CAMELLIA_256_CBC_SHA384 = 0xC095
    TLS_DHE_PSK_WITH_CAMELLIA_128_CBC_SHA256 = 0xC096
    TLS_DHE_PSK_WITH_CAMELLIA_256_CBC_SHA384 = 0xC097
    TLS_RSA_PSK_WITH_CAMELLIA_128_CBC_SHA256 = 0xC098
    TLS_RSA_PSK_WITH_CAMELLIA_256_CBC_SHA384 = 0xC099
    TLS_ECDHE_PSK_WITH_CAMELLIA_128_CBC_SHA256 = 0xC09A
    TLS_ECDHE_PSK_WITH_CAMELLIA_256_CBC_SHA384 = 0xC09B
    TLS_RSA_WITH_AES_128_CCM = 0xC09C
    TLS_RSA_WITH_AES_256_CCM = 0xC09D
    TLS_DHE_RSA_WITH_AES_128_CCM = 0xC09E
    TLS_DHE_RSA_WITH_AES_256_CCM = 0xC09F
    TLS_RSA_WITH_AES_128_CCM_8 = 0xC0A0
    TLS_RSA_WITH_AES_256_CCM_8 = 0xC0A1
    TLS_DHE_RSA_WITH_AES_128_CCM_8 = 0xC0A2
    TLS_DHE_RSA_WITH_AES_256_CCM_8 = 0xC0A3
    TLS_PSK_WITH_AES_128_CCM = 0xC0A4
    TLS_PSK_WITH_AES_256_CCM = 0xC0A5
    TLS_DHE_PSK_WITH_AES_128_CCM = 0xC0A6
    TLS_DHE_PSK_WITH_AES_256_CCM = 0xC0A7
    TLS_PSK_WITH_AES_128_CCM_8 = 0xC0A8
    TLS_PSK_WITH_AES_256_CCM_8 = 0xC0A9
    TLS_PSK_DHE_WITH_AES_128_CCM_8 = 0xC0AA
    TLS_PSK_DHE_WITH_AES_256_CCM_8 = 0xC0AB
    TLS_ECDHE_ECDSA_WITH_AES_128_CCM = 0xC0AC
    TLS_ECDHE_ECDSA_WITH_AES_256_CCM = 0xC0AD
    TLS_ECDHE_ECDSA_WITH_AES_128_CCM_8 = 0xC0AE
    TLS_ECDHE_ECDSA_WITH_AES_256_CCM_8 = 0xC0AF
    TLS_ECCPWD_WITH_AES_128_GCM_SHA256 = 0xC0B0
    TLS_ECCPWD_WITH_AES_256_GCM_SHA384 = 0xC0B1
    TLS_ECCPWD_WITH_AES_128_CCM_SHA256 = 0xC0B2
    TLS_ECCPWD_WITH_AES_256_CCM_SHA384 = 0xC0B3
    TLS_SHA256_SHA256 = 0xC0B4
    TLS_SHA384_SHA384 = 0xC0B5
    TLS_GOSTR341112_256_WITH_KUZNYECHIK_CTR_OMAC = 0xC100
    TLS_GOSTR341112_256_WITH_MAGMA_CTR_OMAC = 0xC101
    TLS_GOSTR341112_256_WITH_28147_CNT_IMIT = 0xC102
    TLS_GOSTR341112_256_WITH_KUZNYECHIK_MGM_L = 0xC103
    TLS_GOSTR341112_256_WITH_MAGMA_MGM_L = 0xC104
    TLS_GOSTR341112_256_WITH_KUZNYECHIK_MGM_S = 0xC105
    TLS_GOSTR341112_256_WITH_MAGMA_MGM_S = 0xC106
    TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256 = 0xCCA8
    TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305_SHA256 = 0xCCA9
    TLS_DHE_RSA_WITH_CHACHA20_POLY1305_SHA256 = 0xCCAA
    TLS_PSK_WITH_CHACHA20_POLY1305_SHA256 = 0xCCAB
    TLS_ECDHE_PSK_WITH_CHACHA20_POLY1305_SHA256 = 0xCCAC
    TLS_DHE_PSK_WITH_CHACHA20_POLY1305_SHA256 = 0xCCAD
    TLS_RSA_PSK_WITH_CHACHA20_POLY1305_SHA256 = 0xCCAE
    TLS_ECDHE_PSK_WITH_AES_128_GCM_SHA256 = 0xD001
    TLS_ECDHE_PSK_WITH_AES_256_GCM_SHA384 = 0xD002
    TLS_ECDHE_PSK_WITH_AES_128_CCM_8_SHA256 = 0xD003
    TLS_ECDHE_PSK_WITH_AES_128_CCM_SHA256 = 0xD005

    @classmethod
    def _missing_(cls, value: object) -> typing.Optional[enum.Enum]:
        return _add_missing_enum_member(cls, value, "Unknown Cipher Suite 0x{0:04X}")


class TlsCompressionMethod(enum.IntEnum):
    none = 0
    deflate = 1
    lzs = 64

    @classmethod
    def _missing_(cls, value: object) -> typing.Optional[enum.Enum]:
        return _add_missing_enum_member(cls, value, "Unknown Compression Method 0x{0:02X}")


class TlsExtensionType(enum.IntEnum):
    server_name = 0
    max_fragment_length = 1
    client_certificate_url = 2
    trusted_ca_keys = 3
    truncated_hmac = 4
    status_request = 5
    user_mapping = 6
    client_authz = 7
    server_authz = 8
    cert_type = 9
    supported_groups = 10
    ec_point_formats = 11
    srp = 12
    signature_algorithms = 13
    use_srtp = 14
    heartbeat = 15
    application_layer_protocol_negotiation = 16
    status_request_v2 = 17
    signed_certificate_timestamp = 18
    client_certificate_type = 19
    server_certificate_type = 20
    padding = 21
    encrypt_then_mac = 22
    extended_master_secret = 23
    token_binding = 24
    cached_info = 25
    tls_lts = 26
    compress_certificate = 27
    record_size_limit = 28
    pwd_protect = 29
    pwd_clear = 30
    password_salt = 31
    ticket_pinning = 32
    tls_cert_with_extern_psk = 33
    delegated_credentials = 34
    session_ticket = 35
    TLMSP = 36
    TLMSP_proxying = 37
    TLMSP_delegate = 38
    supported_ekt_ciphers = 39
    pre_shared_key = 41
    early_data = 42
    supported_versions = 43
    cookie = 44
    psk_key_exchange_modes = 45
    certificate_authorities = 47
    oid_filters = 48
    post_handshake_auth = 49
    signature_algorithms_cert = 50
    key_share = 51
    transparency_info = 52
    connection_id_deprecated = 53
    connection_id = 54
    external_id_hash = 55
    external_session_id = 56
    quic_transport_parameters = 57
    ticket_request = 58
    dnssec_chain = 69
    renegotiation_info = 65281

    @classmethod
    def _missing_(cls, value: object) -> typing.Optional[enum.Enum]:
        return _add_missing_enum_member(cls, value, "Unknown Extension Type 0x{0:04X}")


class TlsServerNameType(enum.IntEnum):
    server_name = 0

    @classmethod
    def _missing_(cls, value: object) -> typing.Optional[enum.Enum]:
        return _add_missing_enum_member(cls, value, "Unknown Server Name Type 0x{0:02X}")


class TlsECPointFormat(enum.IntEnum):
    uncompressed = 0
    ansiX962_compressed_prime = 1
    ansiX962_compressed_char2 = 2

    @classmethod
    def _missing_(cls, value: object) -> typing.Optional[enum.Enum]:
        return _add_missing_enum_member(cls, value, "Unknown EC Point Format 0x{0:02X}")


class TlsSupportedGroup(enum.IntEnum):
    reserved = 0
    sect163k1 = 1
    sect163r1 = 2
    sect163r2 = 3
    sect193r1 = 4
    sect193r2 = 5
    sect233k1 = 6
    sect233r1 = 7
    sect239k1 = 8
    sect283k1 = 9
    sect283r1 = 10
    sect409k1 = 11
    sect409r1 = 12
    sect571k1 = 13
    sect571r1 = 14
    secp160k1 = 15
    secp160r1 = 16
    secp160r2 = 17
    secp192k1 = 18
    secp192r1 = 19
    secp224k1 = 20
    secp224r1 = 21
    secp256k1 = 22
    secp256r1 = 23
    secp384r1 = 24
    secp521r1 = 25
    brainpoolP256r1 = 26
    brainpoolP384r1 = 27
    brainpoolP512r1 = 28
    x25519 = 29
    x448 = 30
    brainpoolP256r1tls13 = 31
    brainpoolP384r1tls13 = 32
    brainpoolP512r1tls13 = 33
    GC256A = 34
    GC256B = 35
    GC256C = 36
    GC256D = 37
    GC512A = 38
    GC512B = 39
    GC512C = 40
    curveSM2 = 41
    ffdhe2048 = 256
    ffdhe3072 = 257
    ffdhe4096 = 258
    ffdhe6144 = 259
    ffdhe8192 = 260
    arbitrary_explicit_prime_curves = 65281
    arbitrary_explicit_char2_curves = 65282

    @classmethod
    def _missing_(cls, value: object) -> typing.Optional[enum.Enum]:
        return _add_missing_enum_member(cls, value, "Unknown Supported Group 0x{0:04X}")


class TlsSignatureScheme(enum.IntEnum):
    rsa_pkcs1_sha1 = 0x0201
    dsa_sha1 = 0x0202
    ecdsa_sha1 = 0x0203
    sha224_rsa = 0x0301
    dsa_sha224 = 0x0302
    sha224_ecdsa = 0x0303
    rsa_pkcs1_sha256 = 0x0401
    dsa_sha256 = 0x0402
    ecdsa_secp256r1_sha256 = 0x0403
    rsa_pkcs1_sha256_legacy = 0x0420
    rsa_pkcs1_sha384 = 0x0501
    dsa_sha384 = 0x0502
    ecdsa_secp384r1_sha384 = 0x0503
    rsa_pkcs1_sha384_legacy = 0x0520
    rsa_pkcs1_sha512 = 0x0601
    dsa_sha512 = 0x0602
    ecdsa_secp521r1_sha512 = 0x0603
    rsa_pkcs1_sha512_legacy = 0x0620
    eccsi_sha256 = 0x0704
    iso_ibs1 = 0x0705
    iso_ibs2 = 0x0706
    iso_chinese_ibs = 0x0707
    sm2sig_sm3 = 0x0708
    gostr34102012_256a = 0x0709
    gostr34102012_256b = 0x070A
    gostr34102012_256c = 0x070B
    gostr34102012_256d = 0x070C
    gostr34102012_512a = 0x070D
    gostr34102012_512b = 0x070E
    gostr34102012_512c = 0x070F
    rsa_pss_rsae_sha256 = 0x0804
    rsa_pss_rsae_sha384 = 0x0805
    rsa_pss_rsae_sha512 = 0x0806
    ed25519 = 0x0807
    ed448 = 0x0808
    rsa_pss_pss_sha256 = 0x0809
    rsa_pss_pss_sha384 = 0x080A
    rsa_pss_pss_sha512 = 0x080B
    ecdsa_brainpoolP256r1tls13_sha256 = 0x081A
    ecdsa_brainpoolP384r1tls13_sha384 = 0x081B
    ecdsa_brainpoolP512r1tls13_sha512 = 0x081C

    @classmethod
    def _missing_(cls, value: object) -> typing.Optional[enum.Enum]:
        return _add_missing_enum_member(cls, value, "Unknown Signature Scheme 0x{0:04X}")


class TlsPskKeyExchangeMode(enum.IntEnum):
    psk_ke = 0
    psk_dhe_ke = 1

    @classmethod
    def _missing_(cls, value: object) -> typing.Optional[enum.Enum]:
        return _add_missing_enum_member(cls, value, "Unknown PSK Key Exchange Mode 0x{0:02X}")


class TlsECCurveType(enum.IntEnum):
    unassigned = 0
    explicit_primve = 1
    explicit_char2 = 2
    named_curve = 3

    @classmethod
    def _missing_(cls, value: object) -> typing.Optional[enum.Enum]:
        return _add_missing_enum_member(cls, value, "Unknown EC Curve Type 0x{0:02X}")


class TlsClientCertificateType(enum.IntEnum):
    rsa_sign = 1
    dss_sign = 2
    rsa_fixed_dh = 3
    dss_fixed_dh = 4
    rsa_ephemeral_dh = 5
    dss_ephemeral_dh = 6
    fortezza_dms = 20
    ecdsa_sign = 64
    rsa_fixed_ecdh = 65
    ecdsa_fixed_ecdh = 66
    gost_sign256 = 67
    gost_sign512 = 68

    @classmethod
    def _missing_(cls, value: object) -> typing.Optional[enum.Enum]:
        return _add_missing_enum_member(cls, value, "Unknown Client Certificate Type 0x{0:02X}")


class DistinguishedNameType(str, enum.Enum):
    object_class = "2.5.4.0"
    aliased_entry_name = "2.5.4.1"
    knowledge_information = "2.5.4.2"
    common_name = "2.5.4.3"
    surname = "2.5.4.4"
    serial_number = "2.5.4.5"
    country_name = "2.5.4.6"
    locality_name = "2.5.4.7"
    state_or_province_name = "2.5.4.8"
    street_address = "2.5.4.9"
    organizational_name = "2.5.4.10"
    organizational_unit_name = "2.5.4.11"
    id_title = "2.5.4.12"
    description = "2.5.4.13"
    search_guide = "2.5.4.14"
    business_category = "2.5.4.15"
    postal_address = "2.5.4.16"
    postal_code = "2.5.4.17"
    post_office_box = "2.5.4.18"
    physical_delivery_office_name = "2.5.4.19"
    telephone_number = "2.5.4.20"
    telex_number = "2.5.4.21"
    teletex_terminal_identifier = "2.5.4.22"
    facsimile_telephone_number = "2.5.4.23"
    x121_address = "2.5.4.24"
    international_isdn_number = "2.5.4.25"
    registered_address = "2.5.4.26"
    destination_indicator = "2.5.4.27"
    preferred_delivery_method = "2.5.4.28"
    presentation_address = "2.5.4.29"
    supported_application_context = "2.5.4.30"
    member = "2.5.4.31"
    owner = "2.5.4.32"
    role_occupant = "2.5.4.33"
    see_also = "2.5.4.34"
    user_password = "2.5.4.35"
    user_certificate = "2.5.4.36"
    ca_certificate = "2.5.4.37"
    authority_revocation_list = "2.5.4.38"
    certificate_revocation_list = "2.5.4.39"
    cross_certificate_pair = "2.5.4.40"
    id_name = "2.5.4.41"
    given_name = "2.5.4.42"
    initials = "2.5.4.43"
    generation_qualifier = "2.5.4.44"
    unique_identifier = "2.5.4.45"
    dn_qualifier = "2.5.4.46"
    enhanced_search_guide = "2.5.4.47"
    protocol_information = "2.5.4.48"
    distinguished_name = "2.5.4.49"
    unique_member = "2.5.4.50"
    house_identifier = "2.5.4.51"
    supported_algorithms = "2.5.4.52"
    delta_revocation_list = "2.5.4.53"
    attribute_certificate = "2.5.4.58"
    pseudonym = "2.5.4.65"

    @classmethod
    def _missing_(cls, value: object) -> typing.Optional[enum.Enum]:
        new_member = str.__new__(cls)
        new_member._name_ = f"Unknown DN OID Type {value}"
        new_member._value_ = str(value)
        return cls._value2member_map_.setdefault(value, new_member)

    @classmethod
    def native_labels(cls) -> typing.Dict["DistinguishedNameType", str]:
        return {
            DistinguishedNameType.object_class: "id-at-objectClass",
            DistinguishedNameType.aliased_entry_name: "id-at-aliasedEntryName",
            DistinguishedNameType.knowledge_information: "id-at-knowldgeinformation",
            DistinguishedNameType.common_name: "id-at-commonName",
            DistinguishedNameType.surname: "id-at-surname",
            DistinguishedNameType.serial_number: "id-at-serialNumber",
            DistinguishedNameType.country_name: "id-at-countryName",
            DistinguishedNameType.locality_name: "id-at-localityName",
            DistinguishedNameType.state_or_province_name: "id-at-stateOrProvinceName",
            DistinguishedNameType.street_address: "id-at-streetAddress",
            DistinguishedNameType.organizational_name: "id-at-organizationName",
            DistinguishedNameType.organizational_unit_name: "id-at-organizationalUnitName",
            DistinguishedNameType.id_title: "id-at-title",
            DistinguishedNameType.description: "id-at-description",
            DistinguishedNameType.search_guide: "id-at-searchGuide",
            DistinguishedNameType.business_category: "id-at-businessCategory",
            DistinguishedNameType.postal_address: "id-at-postalAddress",
            DistinguishedNameType.postal_code: "id-at-postalCode",
            DistinguishedNameType.post_office_box: "id-at-postOfficeBox",
            DistinguishedNameType.physical_delivery_office_name: "id-at-physicalDeliveryOfficeName",
            DistinguishedNameType.telephone_number: "id-at-telephoneNumber",
            DistinguishedNameType.telex_number: "id-at-telexNumber",
            DistinguishedNameType.teletex_terminal_identifier: "id-at-teletexTerminalIdentifier",
            DistinguishedNameType.facsimile_telephone_number: "id-at-facsimileTelephoneNumber",
            DistinguishedNameType.x121_address: "id-at-x121Address",
            DistinguishedNameType.international_isdn_number: "id-at-internationalISDNNumber",
            DistinguishedNameType.registered_address: "id-at-registeredAddress",
            DistinguishedNameType.destination_indicator: "id-at-destinationIndicator",
            DistinguishedNameType.preferred_delivery_method: "id-at-preferredDeliveryMethod",
            DistinguishedNameType.presentation_address: "id-at-presentationAddress",
            DistinguishedNameType.supported_application_context: "id-at-supportedApplicationContext",
            DistinguishedNameType.member: "id-at-member",
            DistinguishedNameType.owner: "id-at-owner",
            DistinguishedNameType.role_occupant: "id-at-roleOccupant",
            DistinguishedNameType.see_also: "id-at-seeAlso",
            DistinguishedNameType.user_password: "id-at-userPassword",
            DistinguishedNameType.user_certificate: "id-at-userCertificate",
            DistinguishedNameType.ca_certificate: "id-at-cACertificate",
            DistinguishedNameType.authority_revocation_list: "id-at-authorityRevocationList",
            DistinguishedNameType.certificate_revocation_list: "id-at-certificateRevocationList",
            DistinguishedNameType.cross_certificate_pair: "id-at-crossCertificatePair",
            DistinguishedNameType.id_name: "id-at-name",
            DistinguishedNameType.given_name: "id-at-givenName",
            DistinguishedNameType.initials: "id-at-initials",
            DistinguishedNameType.generation_qualifier: "id-at-generationQualifier",
            DistinguishedNameType.unique_identifier: "id-at-uniqueIdentifier",
            DistinguishedNameType.dn_qualifier: "id-at-dnQualifier",
            DistinguishedNameType.enhanced_search_guide: "id-at-enhancedSearchGuide",
            DistinguishedNameType.protocol_information: "id-at-protocolInformation",
            DistinguishedNameType.distinguished_name: "id-at-distinguishedName",
            DistinguishedNameType.unique_member: "id-at-uniqueMember",
            DistinguishedNameType.house_identifier: "id-at-houseIdentifier",
            DistinguishedNameType.supported_algorithms: "id-at-supportedAlgorithms",
            DistinguishedNameType.delta_revocation_list: "id-at-deltaRevocationList",
            DistinguishedNameType.attribute_certificate: "id-at-attributeCertificate",
            DistinguishedNameType.pseudonym: "id-at-pseudonym",
        }



# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import base64
import collections.abc
import logging
import os
import typing as t

from spnego._context import (
    IOV,
    ContextProxy,
    ContextReq,
    IOVUnwrapResult,
    IOVWrapResult,
    SecPkgContextSizes,
    UnwrapResult,
    WinRMWrapResult,
    WrapResult,
    split_username,
    wrap_system_error,
)
from spnego._credential import (
    Credential,
    CredentialCache,
    KerberosKeytab,
    Password,
    unify_credentials,
)
from spnego.channel_bindings import GssChannelBindings
from spnego.exceptions import (
    InvalidCredentialError,
    NegotiateOptions,
    NoContextError,
    SpnegoError,
)
from spnego.exceptions import WinError as NativeError
from spnego.iov import BufferType, IOVBuffer, IOVResBuffer

log = logging.getLogger(__name__)

if os.name == "nt":
    import sspilib

    HAS_SSPI = True
else:
    HAS_SSPI = False


def _available_protocols() -> list[str]:
    """Return a list of protocols that SSPIProxy can offer."""
    if HAS_SSPI:
        return ["kerberos", "negotiate", "ntlm"]
    else:
        return []


def _create_iov_result(iov: sspilib.raw.SecBufferDesc) -> tuple[IOVResBuffer, ...]:
    """Converts SSPI IOV buffer to generic IOVBuffer result."""
    buffers = []
    for i in iov:
        buffer_type = int(i.buffer_type)
        if i.buffer_flags & sspilib.raw.SecBufferFlags.SECBUFFER_READONLY_WITH_CHECKSUM:
            buffer_type = BufferType.sign_only
        elif i.buffer_flags & sspilib.raw.SecBufferFlags.SECBUFFER_READONLY:
            buffer_type = BufferType.data_readonly

        buffer_entry = IOVResBuffer(type=BufferType(buffer_type), data=i.data)
        buffers.append(buffer_entry)

    return tuple(buffers)


def _get_sspi_credential(
    principal: str | None,
    protocol: str,
    usage: str,
    credentials: list[Credential],
) -> sspilib.raw.CredHandle:
    """Get the SSPI credential.

    Will get an SSPI credential for the protocol specified. Currently only
    supports Password or CredentialCache credential types.

    Args:
        principal: The principal to use for the AcquireCredentialsHandle call
        protocol: The protocol of the credential.
        usage: Either `initiate` for a client context or `accept` for a server
            context.
        credentials: List of credentials to retrieve from.

    Returns:
        sspilib.raw.CredHandle: The handle to the SSPI credential to use.
    """
    credential_kwargs: dict[str, t.Any] = {
        "package": protocol,
        "principal": principal,
        "credential_use": (
            sspilib.raw.CredentialUse.SECPKG_CRED_OUTBOUND
            if usage == "initiate"
            else sspilib.raw.CredentialUse.SECPKG_CRED_INBOUND
        ),
    }

    for cred in credentials:
        if isinstance(cred, Password):
            domain, username = split_username(cred.username)
            pass_data = sspilib.raw.WinNTAuthIdentity(
                username=username,
                domain=domain,
                password=cred.password,
            )

            return sspilib.raw.acquire_credentials_handle(**credential_kwargs, auth_data=pass_data).credential

        elif isinstance(cred, CredentialCache):
            return sspilib.raw.acquire_credentials_handle(**credential_kwargs).credential

        elif isinstance(cred, KerberosKeytab):
            if not cred.principal:
                raise InvalidCredentialError(context_msg="KerberosKeytab for SSPI requires a principal to be set")

            domain, username = split_username(cred.principal)
            with open(cred.keytab, mode="rb") as kt_file:
                keytab = kt_file.read()

            kt_data = sspilib.raw.WinNTAuthIdentityPackedCredential(
                credential_type=sspilib.raw.WinNTAuthCredentialType.SEC_WINNT_AUTH_DATA_TYPE_KEYTAB,
                credential=keytab,
                username=username,
                domain=domain,
            )

            return sspilib.raw.acquire_credentials_handle(**credential_kwargs, auth_data=kt_data).credential

    raise InvalidCredentialError(context_msg="No applicable credentials available")


class SSPIProxy(ContextProxy):
    """SSPI proxy class for pure SSPI on Windows.

    This proxy class for SSPI exposes this library into a common interface for SPNEGO authentication. This context
    uses compiled C code to interface directly into the SSPI functions on Windows to provide a native SPNEGO
    implementation.
    """

    def __init__(
        self,
        username: str | Credential | list[Credential] | None = None,
        password: str | None = None,
        hostname: str | None = None,
        service: str | None = None,
        channel_bindings: GssChannelBindings | None = None,
        context_req: ContextReq = ContextReq.default,
        usage: str = "initiate",
        protocol: str = "negotiate",
        options: NegotiateOptions = NegotiateOptions.none,
        **kwargs: t.Any,
    ) -> None:

        if not HAS_SSPI:
            raise ImportError("SSPIProxy requires the Windows only sspilib python package")

        credentials = unify_credentials(username, password)
        super(SSPIProxy, self).__init__(
            credentials, hostname, service, channel_bindings, context_req, usage, protocol, options
        )

        self._native_channel_bindings: sspilib.SecChannelBindings | None
        if channel_bindings:
            self._native_channel_bindings = self._get_native_bindings(channel_bindings)
        else:
            self._native_channel_bindings = None

        self._block_size = 0
        self._max_signature = 0
        self._security_trailer = 0

        self._complete = False
        self._context: sspilib.raw.CtxtHandle | None = None
        self.__seq_num = 0

        sspi_credential = kwargs.get("_sspi_credential", None)
        if not sspi_credential:
            try:
                principal = self.spn if usage == "accept" else None
                sspi_credential = _get_sspi_credential(principal, protocol, usage, credentials)
            except NativeError as win_err:
                raise SpnegoError(base_error=win_err, context_msg="Getting SSPI credential") from win_err

        self._credential = sspi_credential

    @classmethod
    def available_protocols(
        cls,
        options: NegotiateOptions | None = None,
    ) -> list[str]:
        return _available_protocols()

    @property
    def client_principal(self) -> str | None:
        if self.usage == "accept":
            names = sspilib.raw.query_context_attributes(
                t.cast(sspilib.raw.CtxtHandle, self._context),
                sspilib.raw.SecPkgContextNames,
            )
            return names.username
        else:
            return None

    @property
    def complete(self) -> bool:
        return self._complete

    @property
    def negotiated_protocol(self) -> str | None:
        # FIXME: Try and replicate GSSAPI. Will return None for acceptor until the first token is returned. Negotiate
        # for both iniator and acceptor until the context is established.
        package_info = sspilib.raw.query_context_attributes(
            t.cast(sspilib.raw.CtxtHandle, self._context),
            sspilib.raw.SecPkgContextPackageInfo,
        )
        return package_info.name.lower()

    @property
    @wrap_system_error(NativeError, "Retrieving session key")
    def session_key(self) -> bytes:
        session_key = sspilib.raw.query_context_attributes(
            t.cast(sspilib.raw.CtxtHandle, self._context),
            sspilib.raw.SecPkgContextSessionKey,
        )
        return session_key.session_key

    def new_context(self) -> SSPIProxy:
        return SSPIProxy(
            hostname=self._hostname,
            service=self._service,
            channel_bindings=self.channel_bindings,
            context_req=self.context_req,
            usage=self.usage,
            protocol=self.protocol,
            options=self.options,
            _sspi_credential=self._credential,
        )

    @wrap_system_error(NativeError, "Processing security token")
    def step(
        self,
        in_token: bytes | None = None,
        *,
        channel_bindings: GssChannelBindings | None = None,
    ) -> bytes | None:
        if not self._is_wrapped:
            log.debug("SSPI step input: %s", base64.b64encode(in_token or b"").decode())

        sec_tokens: list[sspilib.raw.SecBuffer] = []
        if in_token:
            in_token = bytearray(in_token)
            sec_tokens.append(sspilib.raw.SecBuffer(in_token, sspilib.raw.SecBufferType.SECBUFFER_TOKEN))

        native_channel_bindings: sspilib.SecChannelBindings | None
        if channel_bindings:
            native_channel_bindings = self._get_native_bindings(channel_bindings)
        else:
            native_channel_bindings = self._native_channel_bindings

        if native_channel_bindings:
            sec_tokens.append(native_channel_bindings.dangerous_get_sec_buffer())

        in_buffer: sspilib.raw.SecBufferDesc | None = None
        if sec_tokens:
            in_buffer = sspilib.raw.SecBufferDesc(sec_tokens)

        out_buffer = sspilib.raw.SecBufferDesc(
            [
                sspilib.raw.SecBuffer(None, sspilib.raw.SecBufferType.SECBUFFER_TOKEN),
            ]
        )

        context_req: int
        res: sspilib.raw.InitializeContextResult | sspilib.raw.AcceptContextResult
        if self.usage == "initiate":
            context_req = self._context_req | sspilib.IscReq.ISC_REQ_ALLOCATE_MEMORY
            res = sspilib.raw.initialize_security_context(
                credential=self._credential,
                context=self._context,
                target_name=self.spn or "",
                context_req=context_req,
                target_data_rep=sspilib.raw.TargetDataRep.SECURITY_NATIVE_DREP,
                input_buffers=in_buffer,
                output_buffers=out_buffer,
            )
            status = res.status
            self._context = res.context
        else:
            context_req = self._context_req | sspilib.AscReq.ASC_REQ_ALLOCATE_MEMORY
            res = sspilib.raw.accept_security_context(
                credential=self._credential,
                context=self._context,
                input_buffers=in_buffer,
                context_req=context_req,
                target_data_rep=sspilib.raw.TargetDataRep.SECURITY_NATIVE_DREP,
                output_buffers=out_buffer,
            )
            status = res.status
            self._context = res.context

        out_token = out_buffer[0].data or None

        self._context_attr = int(res.attributes)

        if status == sspilib.raw.NtStatus.SEC_E_OK:
            self._complete = True

            attr_sizes = sspilib.raw.query_context_attributes(self._context, sspilib.raw.SecPkgContextSizes)
            self._block_size = attr_sizes.block_size
            self._max_signature = attr_sizes.max_signature
            self._security_trailer = attr_sizes.security_trailer

        if not self._is_wrapped:
            log.debug("SSPI step output: %s", base64.b64encode(out_token or b"").decode())

        return out_token

    def query_message_sizes(self) -> SecPkgContextSizes:
        if not self._security_trailer:
            raise NoContextError(context_msg="Cannot get message sizes until context has been established")

        return SecPkgContextSizes(header=self._security_trailer)

    def wrap(
        self,
        data: bytes,
        encrypt: bool = True,
        qop: int | None = None,
    ) -> WrapResult:
        res = self.wrap_iov([BufferType.header, data, BufferType.padding], encrypt=encrypt, qop=qop)
        return WrapResult(data=b"".join([r.data for r in res.buffers if r.data]), encrypted=res.encrypted)

    @wrap_system_error(NativeError, "Wrapping IOV buffer")
    def wrap_iov(
        self,
        iov: collections.abc.Iterable[IOV],
        encrypt: bool = True,
        qop: int | None = None,
    ) -> IOVWrapResult:
        qop = qop or 0
        if encrypt and qop & sspilib.raw.QopFlags.SECQOP_WRAP_NO_ENCRYPT:
            raise ValueError("Cannot set qop with SECQOP_WRAP_NO_ENCRYPT and encrypt=True")
        elif not encrypt:
            qop |= sspilib.raw.QopFlags.SECQOP_WRAP_NO_ENCRYPT

        buffers = self._build_iov_list(iov, self._convert_iov_buffer)
        iov_buffer = sspilib.raw.SecBufferDesc(buffers)

        sspilib.raw.encrypt_message(
            t.cast(sspilib.raw.CtxtHandle, self._context),
            qop=qop,
            message=iov_buffer,
            seq_no=self._seq_num,
        )

        return IOVWrapResult(buffers=_create_iov_result(iov_buffer), encrypted=encrypt)

    def wrap_winrm(self, data: bytes) -> WinRMWrapResult:
        iov = self.wrap_iov([BufferType.header, data]).buffers
        header = iov[0].data or b""
        enc_data = iov[1].data or b""

        return WinRMWrapResult(header=header, data=enc_data, padding_length=0)

    def unwrap(self, data: bytes) -> UnwrapResult:
        res = self.unwrap_iov([(BufferType.stream, data), BufferType.data])

        dec_data = res.buffers[1].data or b""
        return UnwrapResult(data=dec_data, encrypted=res.encrypted, qop=res.qop)

    @wrap_system_error(NativeError, "Unwrapping IOV buffer")
    def unwrap_iov(
        self,
        iov: collections.abc.Iterable[IOV],
    ) -> IOVUnwrapResult:
        buffers = self._build_iov_list(iov, self._convert_iov_buffer)
        iov_buffer = sspilib.raw.SecBufferDesc(buffers)

        qop = sspilib.raw.decrypt_message(
            t.cast(sspilib.raw.CtxtHandle, self._context),
            iov_buffer,
            seq_no=self._seq_num,
        )
        encrypted = qop & sspilib.raw.QopFlags.SECQOP_WRAP_NO_ENCRYPT == 0

        return IOVUnwrapResult(buffers=_create_iov_result(iov_buffer), encrypted=encrypted, qop=qop)

    def unwrap_winrm(self, header: bytes, data: bytes) -> bytes:
        iov = self.unwrap_iov([(BufferType.header, header), data]).buffers
        return iov[1].data or b""

    @wrap_system_error(NativeError, "Signing message")
    def sign(
        self,
        data: bytes,
        qop: int | None = None,
    ) -> bytes:
        data = bytearray(data)
        signature = bytearray(self._max_signature)
        iov = sspilib.raw.SecBufferDesc(
            [
                sspilib.raw.SecBuffer(data, sspilib.raw.SecBufferType.SECBUFFER_DATA),
                sspilib.raw.SecBuffer(signature, sspilib.raw.SecBufferType.SECBUFFER_TOKEN),
            ]
        )
        sspilib.raw.make_signature(
            t.cast(sspilib.raw.CtxtHandle, self._context),
            qop or 0,
            iov,
            self._seq_num,
        )

        return iov[1].data

    @wrap_system_error(NativeError, "Verifying message")
    def verify(self, data: bytes, mic: bytes) -> int:
        data = bytearray(data)
        mic = bytearray(mic)
        iov = sspilib.raw.SecBufferDesc(
            [
                sspilib.raw.SecBuffer(data, sspilib.raw.SecBufferType.SECBUFFER_DATA),
                sspilib.raw.SecBuffer(mic, sspilib.raw.SecBufferType.SECBUFFER_TOKEN),
            ]
        )

        return sspilib.raw.verify_signature(
            t.cast(sspilib.raw.CtxtHandle, self._context),
            iov,
            self._seq_num,
        )

    @property
    def _context_attr_map(self) -> list[tuple[ContextReq, int]]:
        # The flags values slightly differ for a initiate and accept context.
        attr_map = []

        sspi_req: type[int] | None
        if self.usage == "initiate":
            attr_map.append((ContextReq.no_integrity, "REQ_NO_INTEGRITY"))
            sspi_req = sspilib.IscReq
            sspi_prefix = "ISC"
        else:
            sspi_req = sspilib.AscReq
            sspi_prefix = "ASC"

        attr_map.extend(
            [
                # SSPI does not differ between delegate and delegate_policy, it always respects delegate_policy.
                (ContextReq.delegate, "REQ_DELEGATE"),
                (ContextReq.delegate_policy, "REQ_DELEGATE"),
                (ContextReq.mutual_auth, "REQ_MUTUAL_AUTH"),
                (ContextReq.replay_detect, "REQ_REPLAY_DETECT"),
                (ContextReq.sequence_detect, "REQ_SEQUENCE_DETECT"),
                (ContextReq.confidentiality, "REQ_CONFIDENTIALITY"),
                (ContextReq.integrity, "REQ_INTEGRITY"),
                (ContextReq.dce_style, "REQ_USE_DCE_STYLE"),
                (ContextReq.identify, "REQ_IDENTIFY"),
            ]
        )

        attrs = []
        for spnego_flag, gssapi_name in attr_map:
            attrs.append((spnego_flag, getattr(sspi_req, f"{sspi_prefix}_{gssapi_name}")))

        return attrs

    @property
    def _seq_num(self) -> int:
        num = self.__seq_num
        self.__seq_num += 1
        return num

    def _convert_iov_buffer(self, buffer: IOVBuffer) -> sspilib.raw.SecBuffer:
        data = bytearray()

        if isinstance(buffer.data, bytes):
            data = bytearray(buffer.data)
        elif isinstance(buffer.data, int) and not isinstance(buffer.data, bool):
            data = bytearray(buffer.data)
        else:
            auto_alloc_size = {
                BufferType.header: self._security_trailer,
                BufferType.padding: self._block_size,
                BufferType.trailer: self._security_trailer,
            }

            # If alloc wasn't explicitly set, only alloc if the type is a specific auto alloc type.
            alloc = buffer.data
            if alloc is None:
                alloc = buffer.type in auto_alloc_size

            if alloc:
                if buffer.type not in auto_alloc_size:
                    raise ValueError(
                        "Cannot auto allocate buffer of type %s.%s" % (type(buffer.type).__name__, buffer.type.name)
                    )

                data = bytearray(auto_alloc_size[buffer.type])

        # This buffer types need manual mapping from the generic value to the
        # one understood by sspilib.
        buffer_type = int(buffer.type)
        buffer_flags = 0
        if buffer_type == BufferType.sign_only:
            buffer_type = sspilib.raw.SecBufferType.SECBUFFER_DATA
            buffer_flags = sspilib.raw.SecBufferFlags.SECBUFFER_READONLY_WITH_CHECKSUM
        elif buffer_type == BufferType.data_readonly:
            buffer_type = sspilib.raw.SecBufferType.SECBUFFER_DATA
            buffer_flags = sspilib.raw.SecBufferFlags.SECBUFFER_READONLY

        return sspilib.raw.SecBuffer(data, buffer_type, buffer_flags)

    def _get_native_bindings(
        self,
        channel_bindings: GssChannelBindings,
    ) -> sspilib.SecChannelBindings:
        """Gets the raw byte value of the SEC_CHANNEL_BINDINGS structure."""
        return sspilib.SecChannelBindings(
            initiator_addr_type=int(channel_bindings.initiator_addrtype),
            initiator_addr=channel_bindings.initiator_address,
            acceptor_addr_type=int(channel_bindings.acceptor_addrtype),
            acceptor_addr=channel_bindings.acceptor_address,
            application_data=channel_bindings.application_data,
        )
